'''
Function:
    Implementation of LanZouYParser
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import re
import json
import random
import requests
from urllib.parse import urljoin, urlparse


'''LanZouYParser'''
class LanZouYParser():
    '''parsefromurl'''
    @staticmethod
    def parsefromurl(url: str, passcode: str = '', max_tries: int = 3):
        for _ in range(max_tries):
            try:
                download_result, download_url = LanZouYParser._parsefromurl(url=url, passcode=passcode)
                assert download_url and str(download_url).startswith('http')
                break
            except:
                download_result, download_url = {}, ""
            if not download_url or not str(download_url).startswith('http'):
                file_id = urlparse(url).path.strip('/').split('/')[-1]
                headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/144.0.0.0 Safari/537.36'}
                try:
                    resp = requests.get(f'https://api-v2.cenguigui.cn/api/lanzou/api.php?url=https://cenguigui.lanzouw.com/{file_id}', headers=headers)
                    download_result = resp.json()
                    download_url = download_result['data']['downurl']
                    assert download_url and str(download_url).startswith('http')
                    break
                except:
                    download_result, download_url = {}, ""
        return download_result, download_url
    '''_randip'''
    @staticmethod
    def _randip() -> str:
        ip2 = round(random.randint(600000, 2550000) / 10000)
        ip3 = round(random.randint(600000, 2550000) / 10000)
        ip4 = round(random.randint(600000, 2550000) / 10000)
        arr1 = ["218", "218", "66", "66", "218", "218", "60", "60", "202", "204", "66", "66", "66", "59", "61", "60", "222", "221", "66", "59", "60", "60", "66", "218", "218", "62", "63", "64", "66", "66", "122", "211"]
        ip1 = random.choice(arr1)
        return f"{ip1}.{ip2}.{ip3}.{ip4}"
    '''_httpget'''
    @staticmethod
    def _httpget(url: str, user_agent: str = "", referer: str = "", cookies: dict = None, timeout: int = 10) -> str:
        headers = {"X-FORWARDED-FOR": LanZouYParser._randip(), "CLIENT-IP": LanZouYParser._randip()}
        if user_agent: headers["User-Agent"] = user_agent
        if referer: headers["Referer"] = referer
        resp = requests.get(url, headers=headers, cookies=cookies, timeout=timeout, verify=False, allow_redirects=True)
        resp.raise_for_status()
        resp.encoding = resp.apparent_encoding or "utf-8"
        return resp.text
    '''_httppost'''
    @staticmethod
    def _httppost(data: dict, url: str, referer: str = "", user_agent: str = "", timeout: int = 10) -> str:
        headers = {"X-FORWARDED-FOR": LanZouYParser._randip(), "CLIENT-IP": LanZouYParser._randip()}
        if user_agent: headers["User-Agent"] = user_agent
        if referer: headers["Referer"] = referer
        resp = requests.post(url, data=data, headers=headers, timeout=timeout, verify=False, allow_redirects=True)
        resp.raise_for_status()
        resp.encoding = resp.apparent_encoding or "utf-8"
        return resp.text
    '''_httpredirecturl'''
    @staticmethod
    def _httpredirecturl(url: str, referer: str, user_agent: str, cookie_str: str, timeout: int = 10) -> str:
        headers = {
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8", "Accept-Encoding": "gzip, deflate", 
            "Accept-Language": "zh-CN,zh;q=0.9", "Cache-Control": "no-cache", "Connection": "keep-alive", "Pragma": "no-cache", "Upgrade-Insecure-Requests": "1", 
            "User-Agent": user_agent, "Referer": referer, "Cookie": cookie_str,
        }
        resp = requests.get(url, headers=headers, timeout=timeout, verify=False, allow_redirects=False)
        resp.raise_for_status()
        loc = resp.headers.get("Location", "") or resp.headers.get("location", "")
        if not loc: return ""
        return urljoin(url, loc)
    '''_acwscv2simple'''
    @staticmethod
    def _acwscv2simple(arg1: str):
        if not arg1: return ""
        mask = "3000176000856006061501533003690027800375"
        pos_list = (15, 35, 29, 24, 33, 16, 1, 38, 10, 9, 19, 31, 40, 27, 22, 23, 25, 13, 6, 11, 39, 18, 20, 8, 14, 21, 32, 26, 2, 30, 7, 4, 17, 5, 3, 28, 34, 37, 12, 36)
        arg2 = "".join(arg1[p - 1] for p in pos_list if p <= len(arg1))
        length = min(len(arg2), len(mask))
        return "".join(f"{(int(arg2[i:i+2], 16) ^ int(mask[i:i+2], 16)):02x}" for i in range(0, length, 2))
    '''_parsefromurl'''
    @staticmethod
    def _parsefromurl(url: str, passcode: str = ''):
        # init
        download_result, user_agent = {}, "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/143.0.0.0 Safari/537.36"
        normalize_lanzou_url_func = lambda u: ("https://www.lanzouf.com/" + t.lstrip("/") if (t := (u.split(".com/", 1)[1] if ".com/" in u else None)) is not None else ("https://www.lanzouf.com" + u) if u.startswith("/") else u if u.startswith("http") else "https://www.lanzouf.com/" + u.lstrip("/"))
        extract_first_func = lambda regex_list, text: next((m.group(1) for rgx in regex_list if (m := re.search(rgx, text, flags=re.S))), "")
        # vist home page
        url = normalize_lanzou_url_func(url)
        homepage_url_html = LanZouYParser._httpget(url, user_agent=user_agent)
        if "文件取消分享了" in homepage_url_html: raise
        soft_name = extract_first_func([r'style="font-size: 30px;text-align: center;padding: 56px 0px 20px 0px;">(.*?)</div>', r'<div class="n_box_3fn".*?>(.*?)</div>', r"var filename = '(.*?)';", r'div class="b"><span>(.*?)</span></div>'], homepage_url_html)
        soft_size = extract_first_func([r'<div class="n_filesize".*?>大小：(.*?)</div>', r'<span class="p7">文件大小：</span>(.*?)<br>'], homepage_url_html)
        # with passcode
        if "function down_p(){" in homepage_url_html:
            segment = re.findall(r"'sign':'(.*?)',", homepage_url_html, flags=re.S)
            ajaxm = re.findall(r"ajaxm\.php\?file=\d+", homepage_url_html, flags=re.S)
            assert not (len(segment) < 2 or len(ajaxm) < 1)
            post_data = {"action": "downprocess", "sign": segment[1], "p": passcode, "kd": 1}
            post_url = "https://www.lanzouf.com/" + ajaxm[0]
            parse_result = LanZouYParser._httppost(post_data, post_url, referer=url, user_agent=user_agent)
            parse_result: dict = json.loads(parse_result)
            soft_name = parse_result.get("inf") or soft_name
        # without passcode    
        else:
            link = extract_first_func([r'\n<iframe.*?name="[\s\S]*?"\ssrc="\/(.*?)"', r'<iframe.*?name="[\s\S]*?"\ssrc="\/(.*?)"'], homepage_url_html)
            assert link
            ifurl = "https://www.lanzouf.com/" + link.lstrip("/")
            iframe_html = LanZouYParser._httpget(ifurl, user_agent=user_agent)
            wp_sign = re.findall(r"wp_sign = '(.*?)'", iframe_html, flags=re.S)
            ajaxdata = re.findall(r"ajaxdata = '(.*?)'", iframe_html, flags=re.S)
            ajaxm = re.findall(r"ajaxm\.php\?file=\d+", iframe_html, flags=re.S)
            assert not (len(wp_sign) < 1 or len(ajaxdata) < 1 or len(ajaxm) < 2)
            post_data = {"action": "downprocess", "websignkey": ajaxdata[0], "signs": ajaxdata[0], "sign": wp_sign[0], "websign": "", "kd": 1, "ves": 1}
            post_url = "https://www.lanzouf.com/" + ajaxm[1]
            parse_result = LanZouYParser._httppost(post_data, post_url, referer=ifurl, user_agent=user_agent)
            parse_result: dict = json.loads(parse_result)
        # final parse
        assert not (not isinstance(parse_result, dict) or parse_result.get("zt") != 1)
        download_url = f"{parse_result['dom']}/file/{parse_result['url']}"
        download_html = LanZouYParser._httpget(download_url, user_agent=user_agent)
        arg1_list = re.findall(r"arg1='(.*?)'", download_html, flags=re.S)
        if arg1_list:
            decrypted = LanZouYParser._acwscv2simple(arg1_list[0])
            cookie_str = f"down_ip=1; expires=Sat, 16-Nov-2019 11:42:54 GMT; path=/; domain=.baidupan.com; acw_sc__v2={decrypted}"
            redirected_download_url = LanZouYParser._httpredirecturl(download_url, referer="https://developer.lanzoug.com", user_agent=user_agent, cookie_str=cookie_str)
            if "http" in (redirected_download_url or ""): download_url = redirected_download_url
        download_url = re.sub(r"pid=[^&]*&", "", download_url)
        download_result = {"name": soft_name or "", "filesize": soft_size or "", "downUrl": download_url, "parse_result": parse_result}
        # return
        return download_result, download_url


'''
Function:
    Implementation of FLMP3MusicClient: https://www.flmp3.pro/index.html
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
from bs4 import BeautifulSoup
from .base import BaseMusicClient
from rich.progress import Progress
from urllib.parse import urljoin, urlparse
from ..utils import legalizestring, usesearchheaderscookies, seconds2hms, searchdictbykey, safeextractfromdict, SongInfo, QuarkParser


'''FLMP3MusicClient'''
class FLMP3MusicClient(BaseMusicClient):
    source = 'FLMP3MusicClient'
    MUSIC_QUALITY_RANK = {"DSD": 100, "DSF": 100, "DFF": 100, "WAV": 95, "AIFF": 95, "FLAC": 90, "ALAC": 90, "APE": 88, "WV": 88, "OPUS": 70, "AAC": 65, "M4A": 65, "OGG": 60, "VORBIS": 60, "MP3": 50, "WMA": 45}
    def __init__(self, **kwargs):
        super(FLMP3MusicClient, self).__init__(**kwargs)
        assert self.quark_parser_config.get('cookies'), f'{self.source}.__init__ >>> "quark_parser_config" is not configured, so the songs cannot be downloaded.'
        self.default_search_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36',
        }
        self.default_download_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36',
        }
        self.default_headers = self.default_search_headers
        self._initsession()
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        # construct search urls
        self.search_size_per_page = min(self.search_size_per_source, 12)
        search_urls, page_size, count = [], self.search_size_per_page, 0
        while self.search_size_per_source > count:
            if int(count // page_size) + 1 == 1: search_urls.append(f'https://www.flmp3.pro/search.html?keyword={keyword}')
            else: search_urls.append(f'https://www.flmp3.pro/search.html?page={int(count // page_size) + 1}&keyword={keyword}')
            count += page_size
        # return
        return search_urls
    '''_parsesearchresultsfromhtml'''
    def _parsesearchresultsfromhtml(self, html_text: str):
        soup = BeautifulSoup(html_text, "html.parser")
        search_results, base_url = [], "https://flmp3.pro"
        for li in soup.select("div.list ul.flex.flex-wrap > li"):
            a = li.select_one("a")
            if not a: continue
            song_href = a.get("href", ""); song_url = urljoin(base_url, song_href) if song_href else None; title_el = li.select_one("div.con div.t h3")
            artist_el = li.select_one("div.con div.t p"); date_el = li.select_one("div.con div.date"); img_el = li.select_one("div.pic img")
            search_results.append({"song_url": song_url, "title": title_el.get_text(strip=True) if title_el else None, "artist": artist_el.get_text(strip=True) if artist_el else None, "date": date_el.get_text(strip=True) if date_el else None, "img_url": img_el.get("src") if img_el else None, "img_alt": img_el.get("alt") if img_el else None})
        return search_results
    '''_parsesongdetailfordownloadpages'''
    def _parsesongdetailfordownloadpages(self, html_text: str):
        infer_quality_func = lambda text: next((q for q in FLMP3MusicClient.MUSIC_QUALITY_RANK.keys() if q in str(text).upper()), "UNKNOWN")
        soup, base_url, links = BeautifulSoup(html_text, "html.parser"), "https://www.flmp3.pro", []
        for a in soup.select(".btnBox a[href]"):
            text, href = a.get_text(strip=True), a["href"]
            if not href: continue
            links.append({"text": text, "quality": infer_quality_func(text), "rank": FLMP3MusicClient.MUSIC_QUALITY_RANK.get(infer_quality_func(text), 0), "url": urljoin(base_url, href)})
        links_sorted = sorted(links, key=lambda x: x["rank"], reverse=True)
        song_id = urlparse(str(links_sorted[0]['url'])).path.strip('/').split('/')[-1].split('.')[0]
        return {'links_sorted': links_sorted, 'song_id': song_id}
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        # successful
        try:
            # --search results
            resp = self.get(search_url, **request_overrides)
            resp.raise_for_status()
            search_results = self._parsesearchresultsfromhtml(resp.text)
            for search_result in search_results:
                # --download results
                if not isinstance(search_result, dict) or ('song_url' not in search_result): continue
                song_info = SongInfo(source=self.source)
                try: resp = self.get(search_result['song_url'], **request_overrides); resp.raise_for_status(); download_result = self._parsesongdetailfordownloadpages(resp.text)
                except: continue
                if not download_result['links_sorted']: continue
                for download_page_details in download_result['links_sorted']:
                    try:
                        resp = self.get(download_page_details['url'], **request_overrides)
                        resp.raise_for_status()
                        soup = BeautifulSoup(resp.text, "lxml")
                        quark_download_url = soup.select_one("a.linkbtn[href]")['href']
                    except:
                        continue
                    if not quark_download_url: continue
                    download_result['quark_parse_result'], download_url = QuarkParser.parsefromurl(quark_download_url, **self.quark_parser_config)
                    if not download_url or not str(download_url).startswith('http'): continue
                    duration = [int(float(d)) for d in searchdictbykey(download_result, 'duration') if int(float(d)) > 0]
                    duration_s = duration[0] if duration else 0
                    song_info = SongInfo(
                        raw_data={'search': search_result, 'download': download_result, 'lyric': {}}, source=self.source, song_name=legalizestring(safeextractfromdict(search_result, ['title'], None)),
                        singers=legalizestring(safeextractfromdict(search_result, ['artist'], None)), album='NULL', ext='mp3', file_size='NULL', identifier=download_result['song_id'], duration_s=duration_s,
                        duration=seconds2hms(duration_s), lyric='NULL', cover_url=safeextractfromdict(search_result, ['img_url'], None), download_url=download_url, 
                        download_url_status=self.quark_audio_link_tester.test(download_url, request_overrides), default_download_headers=self.quark_default_download_headers,
                    )
                    song_info.download_url_status['probe_status'] = self.quark_audio_link_tester.probe(song_info.download_url, request_overrides)
                    song_info.file_size = song_info.download_url_status['probe_status']['file_size']
                    song_info.ext = song_info.download_url_status['probe_status']['ext'] if (song_info.download_url_status['probe_status']['ext'] and song_info.download_url_status['probe_status']['ext'] not in ('NULL', )) else song_info.ext
                    if song_info.with_valid_download_url: break
                if not song_info.with_valid_download_url: continue
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos


'''
Function:
    Implementation of AppleMusicClient: https://music.apple.com/{geo}/new
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import os
import copy
import shutil
from types import SimpleNamespace
from .base import BaseMusicClient
from urllib.parse import urlencode
from rich.progress import Progress
from ..utils.appleutils import AppleMusicClientDownloadSongUtils, AppleMusicClientAPIUtils, AppleMusicClientItunesApiUtils, DownloadItem, SongCodec, RemuxMode
from ..utils import touchdir, legalizestring, resp2json, seconds2hms, usesearchheaderscookies, safeextractfromdict, usedownloadheaderscookies, cleanlrc, SongInfo, SongInfoUtils


'''AppleMusicClient'''
class AppleMusicClient(BaseMusicClient):
    source = 'AppleMusicClient'
    def __init__(self, **kwargs):
        use_wrapper, wrapper_account_url, language, codec, wrapper_decrypt_ip = kwargs.pop('use_wrapper', False), kwargs.pop('wrapper_account_url', "http://127.0.0.1:30020/"), kwargs.pop('language', "en-US"), kwargs.pop('codec', None), kwargs.pop('wrapper_decrypt_ip', "127.0.0.1:10020")
        super(AppleMusicClient, self).__init__(**kwargs)
        if (not self.default_cookies or 'media-user-token' not in self.default_cookies) and (not use_wrapper): self.logger_handle.warning(f'{self.source}.__init__ >>> both "media-user-token" and "use_wrapper" are not configured, so song downloads are restricted and only the preview portion of the track can be downloaded.')
        self.apple_music_api, self.itunes_api, self.use_wrapper, self.wrapper_account_url, self.language, self.account_info, self.codec, self.wrapper_decrypt_ip = None, None, use_wrapper, wrapper_account_url, language, {}, codec, wrapper_decrypt_ip
        if self.codec is None and use_wrapper: self.codec = SongCodec.ALAC
        if self.codec is None and (not use_wrapper) and (self.default_cookies and 'media-user-token' in self.default_cookies): self.codec = SongCodec.AAC_LEGACY
        self.default_search_headers = {
            "accept": "*/*", "accept-language": "en-US", "origin": "https://music.apple.com", "priority": "u=1, i", "referer": "https://music.apple.com", "sec-fetch-site": "same-site",
            "sec-ch-ua": '"Google Chrome";v="137", "Chromium";v="137", "Not/A)Brand";v="24"', "sec-ch-ua-mobile": "?0", "sec-ch-ua-platform": '"Windows"', "sec-fetch-mode": "cors", 
            "sec-fetch-dest": "empty", "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/137.0.0.0 Safari/537.36",
        }
        self.default_download_headers = {"user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/137.0.0.0 Safari/537.36"}
        self.default_headers = self.default_search_headers
        self._initsession()
    '''_download'''
    @usedownloadheaderscookies
    def _download(self, song_info: SongInfo, request_overrides: dict = None, downloaded_song_infos: list = [], progress: Progress = None, song_progress_id: int = 0):
        if isinstance(song_info.download_url, str): return super()._download(song_info=song_info, request_overrides=request_overrides, downloaded_song_infos=downloaded_song_infos, progress=progress, song_progress_id=song_progress_id)
        request_overrides = request_overrides or {}
        try:
            touchdir(song_info.work_dir)
            tmp_dir = f'apple_id_{str(song_info.identifier)}'
            touchdir(tmp_dir)
            download_item: DownloadItem = song_info.download_url
            progress.update(song_progress_id, total=1, kind='overall')
            progress.update(song_progress_id, description=f"{self.source}.download >>> {song_info.song_name[:10] + '...' if len(song_info.song_name) > 13 else song_info.song_name[:13]} (Downloading)")
            AppleMusicClientDownloadSongUtils.download(download_item=download_item, work_dir=tmp_dir, silent=self.disable_print, codec=self.codec, wrapper_decrypt_ip=self.wrapper_decrypt_ip, artist=song_info.singers, use_wrapper=self.use_wrapper, remux_mode=RemuxMode.FFMPEG)
            shutil.move(download_item.staged_path, song_info.save_path)
            progress.update(song_progress_id, total=os.path.getsize(song_info.save_path), kind='download')
            progress.advance(song_progress_id, os.path.getsize(song_info.save_path))
            progress.update(song_progress_id, description=f"{self.source}.download >>> {song_info.song_name[:10] + '...' if len(song_info.song_name) > 13 else song_info.song_name[:13]} (Success)")
            downloaded_song_infos.append(SongInfoUtils.fillsongtechinfo(copy.deepcopy(song_info), logger_handle=self.logger_handle, disable_print=self.disable_print))
            shutil.rmtree(tmp_dir, ignore_errors=True)
        except Exception as err:
            progress.update(song_progress_id, description=f"{self.source}.download >>> {song_info.song_name[:10] + '...' if len(song_info.song_name) > 13 else song_info.song_name[:13]} (Error: {err})")
        return downloaded_song_infos
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        # init apple_music_api
        if self.use_wrapper and (not self.apple_music_api): self.apple_music_api = AppleMusicClientAPIUtils.createfromwrapper(wrapper_account_url=self.wrapper_account_url, request_overrides=request_overrides, language=self.language)
        elif self.default_cookies and ('media-user-token' in self.default_cookies) and (not self.apple_music_api): self.apple_music_api = AppleMusicClientAPIUtils.createfromnetscapecookies(cookies=self.default_cookies, request_overrides=request_overrides, language=self.language)
        # init itunes_api
        if self.apple_music_api and (not self.itunes_api): self.itunes_api = AppleMusicClientItunesApiUtils(storefront=self.apple_music_api.storefront, language=self.apple_music_api.language)
        # init default_search_headers
        if self.apple_music_api and ('authorization' not in self.default_headers):
            self.default_search_headers = copy.deepcopy(self.apple_music_api.client.headers)
            self.default_headers = self.default_search_headers
            self._initsession()
            self.account_info = self.apple_music_api.account_info
        elif ('authorization' not in self.default_headers):
            virtual_client = SimpleNamespace(client=self.session, language=self.language)
            self.default_search_headers.update({"authorization": f"Bearer {AppleMusicClientAPIUtils.gettoken(virtual_client, request_overrides=request_overrides)}"})
            self.default_headers = self.default_search_headers
            self._initsession()
        # search rules
        default_rule = {
            "groups": "song", "l": "en-US", "offset": "0", "term": keyword, "types": "activities,albums,apple-curators,artists,curators,editorial-items,music-movies,music-videos,playlists,record-labels,songs,stations,tv-episodes,uploaded-videos",
            "art[url]": "f", "extend": "artistUrl", "fields[albums]": "artistName,artistUrl,artwork,contentRating,editorialArtwork,editorialNotes,name,playParams,releaseDate,url,trackCount", "fields[artists]": "url,name,artwork",
            "format[resources]": "map", "include[editorial-items]": "contents", "include[songs]": "artists", "limit": "10", "omit[resource]": "autos", "platform": "web", "relate[albums]": "artists", "relate[editorial-items]": "contents",
            "relate[songs]": "albums", "types": "activities,albums,apple-curators,artists,curators,music-movies,music-videos,playlists,songs,stations,tv-episodes,uploaded-videos", "with": "lyrics,serverBubbles", 
        }
        default_rule.update(rule)
        geo = safeextractfromdict(self.account_info, ['meta', 'subscription', 'storefront'], 'us')
        # construct search urls based on search rules
        base_url = f'https://amp-api-edge.music.apple.com/v1/catalog/{geo}/search?'
        search_urls, page_size, count = [], self.search_size_per_page, 0
        while self.search_size_per_source > count:
            page_rule = copy.deepcopy(default_rule)
            page_rule['limit'] = page_size
            page_rule['offset'] = count
            search_urls.append(base_url + urlencode(page_rule))
            count += page_size
        # return
        return search_urls
    '''_parsewithnonvipofficialapi'''
    def _parsewithnonvipofficialapi(self, search_result: dict, request_overrides: dict = None):
        # init
        request_overrides, song_info = request_overrides or {}, SongInfo(source=self.source)
        # parse
        download_url: str = safeextractfromdict(search_result, ['attributes', 'previews', 0, 'url'], '')
        if not download_url or not download_url.startswith('http'): return song_info
        song_info = SongInfo(
            raw_data={'search': search_result, 'download': {}, 'lyric': {}}, source=self.source, song_name=legalizestring(safeextractfromdict(search_result, ['attributes', 'name'], None)),
            singers=legalizestring(safeextractfromdict(search_result, ['attributes', 'artistName'], None)), album=legalizestring(safeextractfromdict(search_result, ['attributes', 'albumName'], None)),
            ext=download_url.split('?')[0].split('.')[-1], file_size='NULL', identifier=search_result['id'], duration_s=float(safeextractfromdict(search_result, ['attributes', 'durationInMillis'], 0)) / 1000,
            duration=seconds2hms(float(safeextractfromdict(search_result, ['attributes', 'durationInMillis'], 0)) / 1000), lyric=None, cover_url=safeextractfromdict(search_result, ['attributes', 'artwork', 'url'], ""),
            download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
        )
        if not song_info.with_valid_download_url: return song_info
        if song_info.cover_url and song_info.cover_url.startswith('http'): song_info.cover_url = song_info.cover_url.format(w=600, h=600, f='jpg')
        song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
        song_info.file_size = song_info.download_url_status['probe_status']['file_size']
        song_info.ext = song_info.download_url_status['probe_status']['ext'] if (song_info.download_url_status['probe_status']['ext'] and song_info.download_url_status['probe_status']['ext'] not in ('NULL',)) else song_info.ext
        # return
        return song_info
    '''_parsewithvipofficialapi'''
    def _parsewithvipofficialapi(self, search_result: dict, request_overrides: dict = None):
        # init
        request_overrides, song_info, codec = request_overrides or {}, SongInfo(source=self.source), self.codec
        # parse
        geo = safeextractfromdict(self.account_info, ['meta', 'subscription', 'storefront'], 'us')
        params = {"extend": "extendedAssetUrls", "include": "lyrics,albums"}
        resp = self.get(f'https://amp-api.music.apple.com/v1/catalog/{geo}/songs/{search_result["id"]}', params=params, **request_overrides)
        resp.raise_for_status()
        download_result = resp2json(resp=resp)
        download_item: DownloadItem = AppleMusicClientDownloadSongUtils.getdownloaditem(
            song_metadata=download_result['data'][0], playlist_metadata=None, codec=codec, apple_music_api=self.apple_music_api, itunes_api=self.itunes_api, request_overrides=request_overrides, use_wrapper=self.use_wrapper
        )
        try: resp = self.get(download_item.stream_info.audio_track.stream_url, **request_overrides); resp.raise_for_status()
        except Exception: return song_info
        song_info = SongInfo(
            raw_data={'search': search_result, 'download': download_result, 'lyric': {}}, source=self.source, song_name=legalizestring(safeextractfromdict(search_result, ['attributes', 'name'], None)),
            singers=legalizestring(safeextractfromdict(search_result, ['attributes', 'artistName'], None)), album=legalizestring(safeextractfromdict(search_result, ['attributes', 'albumName'], None)),
            ext=download_item.stream_info.file_format.value, file_size='HLS', identifier=search_result['id'], duration_s=float(safeextractfromdict(search_result, ['attributes', 'durationInMillis'], 0)) / 1000,
            duration=seconds2hms(int(float(safeextractfromdict(search_result, ['attributes', 'durationInMillis'], 0)) / 1000)), lyric=cleanlrc(str(download_item.lyrics.synced)) or 'NULL', 
            cover_url=safeextractfromdict(search_result, ['attributes', 'artwork', 'url'], ""), download_url=download_item, download_url_status={'ok': True},
        )
        if not song_info.with_valid_download_url: return song_info
        if song_info.cover_url and song_info.cover_url.startswith('http'): song_info.cover_url = song_info.cover_url.format(w=600, h=600, f='jpg')
        # return
        return song_info
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        # successful
        try:
            # --search results
            resp = self.get(search_url, **request_overrides)
            resp.raise_for_status()
            search_results: dict = resp2json(resp)['resources']['songs']
            for song_key, search_result in search_results.items():
                # --download results
                if not isinstance(search_result, dict) or ('id' not in search_result) or (search_result.get('type') not in {'songs'}): continue
                search_result['song_key'], song_info = song_key, SongInfo(source=self.source)
                # ----non-vip users
                if (not self.default_cookies or 'media-user-token' not in self.default_cookies) and (not self.use_wrapper):
                    try: song_info = self._parsewithnonvipofficialapi(search_result=search_result, request_overrides=request_overrides)
                    except Exception: continue
                # ----vip users
                else:
                    try: song_info = self._parsewithvipofficialapi(search_result=search_result, request_overrides=request_overrides)
                    except Exception: continue
                if not song_info.with_valid_download_url: continue
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos


'''
Function:
    Implementation of QingtingMusicClient: https://m.qingting.fm/
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import copy
import hmac
import math
import hashlib
from rich.progress import Progress
from typing import Any, Dict, List
from ..sources import BaseMusicClient
from urllib.parse import urlencode, urlparse, parse_qs
from ..utils import legalizestring, resp2json, seconds2hms, usesearchheaderscookies, safeextractfromdict, byte2mb, SongInfo


'''QingtingMusicClient'''
class QingtingMusicClient(BaseMusicClient):
    source = 'QingtingMusicClient'
    HMAC_KEY = "99@b8#571(bb38_b"
    DEVICE_ID = "66f6e3b560ad8876e52e6e67ee535c5c"
    ALLOWED_SEARCH_TYPES = ['album', 'track']
    def __init__(self, **kwargs):
        self.allowed_search_types = list(set(kwargs.pop('allowed_search_types', QingtingMusicClient.ALLOWED_SEARCH_TYPES)))
        super(QingtingMusicClient, self).__init__(**kwargs)
        if self.default_cookies: assert ("qingting_id" in self.default_cookies) and (("access_token" in self.default_cookies) or ("refresh_token" in self.default_cookies)), '"qingting_id", "access_token" and "refresh_token" should be configured, refer to "https://musicdl.readthedocs.io/zh/latest/Quickstart.html#qingtingfm-audio-radio-download"'
        self.default_search_headers = {"User-Agent": "QingTing-iOS/10.7.9.0 com.Qting.QTTour Mozilla/5.0 (iPhone; CPU iPhone OS 16_6_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/15E148", "QT-App-Version": "10.7.9.0"}
        self.default_download_headers = {"User-Agent": "QingTing-iOS/10.7.9.0 com.Qting.QTTour Mozilla/5.0 (iPhone; CPU iPhone OS 16_6_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/15E148", "QT-App-Version": "10.7.9.0"}
        self.default_headers = self.default_search_headers
        self.auth_info = copy.deepcopy(self.default_search_cookies or self.default_download_cookies)
        self.default_search_cookies = {}
        self.default_download_cookies = {}
        self._initsession()
    '''_auth'''
    def _auth(self, request_overrides: dict = None):
        request_overrides = request_overrides or {}
        qingting_id, refresh_token = self.auth_info['qingting_id'], self.auth_info['refresh_token']
        resp = self.post("https://user.qtfm.cn/u2/api/v4/auth", headers={"Content-Type": "application/x-www-form-urlencoded"}, data={"refresh_token": refresh_token, "qingting_id": qingting_id, "device_id": QingtingMusicClient.DEVICE_ID, "grant_type": "refresh_token"}, **request_overrides)
        resp.raise_for_status()
        auth_info = resp2json(resp)['data']
        self.auth_info = copy.deepcopy(auth_info)
        return auth_info
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        if self.auth_info and ("access_token" not in self.auth_info): self._auth()
        # search rules: sort_type should be in {"0", "1", "2"} >>> {Comprehensive Sorting, Most Popular, Latest Updates}; include should be in {"channel_ondemand", "channel_live", "program_ondemand", "people_podcaster", "all"}
        default_rule = {"k": keyword, "sort_type": '0', "page": "1", "include": "channel_ondemand", "pagesize": "30", "k_src": "direct"}
        default_rule.update(rule)
        # construct search urls based on search rules
        base_url = 'https://app.qtfm.cn/m-bff/v1/search/result?'
        search_urls, page_size = [], self.search_size_per_page
        for search_type in QingtingMusicClient.ALLOWED_SEARCH_TYPES:
            if search_type not in self.allowed_search_types: continue
            default_rule_search_type = copy.deepcopy(default_rule)
            default_rule_search_type['include'], count = {"album": "channel_ondemand", "track": "program_ondemand"}[search_type], 0
            while self.search_size_per_source > count:
                page_rule = copy.deepcopy(default_rule_search_type)
                page_rule['pagesize'] = str(page_size)
                page_rule['page'] = str(int(count // page_size) + 1)
                search_urls.append(base_url + urlencode(page_rule))
                count += page_size
        # return
        return search_urls
    '''_fetchchannelinfo'''
    def _fetchchannelinfo(self, channel_id: str, request_overrides: dict = None) -> Dict[str, Any]:
        request_overrides = request_overrides or {}
        url = f"https://app.qtfm.cn/m-bff/v2/channel/{channel_id}"
        resp = self.get(url, **request_overrides)
        resp.raise_for_status()
        channel_info = resp2json(resp=resp)
        return channel_info
    '''_listpageprograms'''
    def _listpageprograms(self, channel_id: str, page: int, page_size: int, request_overrides: dict = None) -> List[Dict[str, Any]]:
        request_overrides = request_overrides or {}
        url = f"https://app.qtfm.cn/m-bff/v2/channel/{channel_id}/programs"
        resp = self.get(url, params={"order": "asc", "pagesize": str(page_size), "curpage": str(page)}, **request_overrides)
        resp.raise_for_status()
        programs = resp2json(resp=resp)
        return programs
    '''_parsewithofficialapiv1'''
    def _parsewithofficialapiv1(self, search_result: dict, request_overrides: dict = None):
        # init
        request_overrides, song_id, song_info, app_url = request_overrides or {}, search_result.get('id') or search_result.get('Id'), SongInfo(source=self.source), search_result.get('url')
        if not song_id or not app_url: return song_info
        hmac_md5_hex_func = lambda key, msg: hmac.new(str(key).encode("utf-8"), str(msg).encode("utf-8"), hashlib.md5).hexdigest()
        # parse
        parsed_app_url_params = parse_qs(urlparse(str(app_url)).query, keep_blank_values=True)
        channel_id, program_id = parsed_app_url_params.get('channel_id')[0], (parsed_app_url_params.get('program_id') or [song_id])[0]
        assert str(song_id) == str(program_id), 'song_id and app_url are not synchronized'
        path_query = f"/m-bff/v1/audiostreams/channel/{channel_id}/program/{program_id}?access_token={self.auth_info.get('access_token', '')}&device_id={QingtingMusicClient.DEVICE_ID}&qingting_id={self.auth_info.get('qingting_id', '')}&type=play"
        sign = hmac_md5_hex_func(QingtingMusicClient.HMAC_KEY, path_query)
        resp = self.get(f"https://app.qtfm.cn{path_query}&sign={sign}", **request_overrides)
        resp.raise_for_status()
        download_result = resp2json(resp=resp)
        if 'channel_info' not in search_result:
            try: search_result['channel_info'] = self._fetchchannelinfo(channel_id, request_overrides)
            except Exception: pass
        candidate_editions: list[dict] = sorted(download_result['data']['editions'] + (download_result['data'].get('backup_editions') if isinstance(download_result['data'].get('backup_editions'), list) else []), key=lambda x: (x.get('size', 0), x.get('bitrate', 0)), reverse=True)
        for edition in candidate_editions:
            if not edition.get('urls'): continue
            if isinstance(edition.get('urls'), str): edition['urls'] = [edition.get('urls')]
            for download_url in edition.get('urls'):
                if not download_url or not str(download_url).startswith('http'): continue
                song_info = SongInfo(
                    raw_data={'search': search_result, 'download': download_result, 'lyric': {}}, source=self.source, song_name=legalizestring(search_result.get('title')), singers=legalizestring(', '.join([singer.get('nick_name') for singer in (safeextractfromdict(search_result, ['channel_info', 'data', 'podcasters'], []) or []) if isinstance(singer, dict) and singer.get('nick_name')])),
                    album=legalizestring(safeextractfromdict(search_result, ['channel_info', 'data', 'title'], None) or search_result.get('desc')), ext=download_url.split('?')[0].split('.')[-1], file_size_bytes=int(float(edition.get('size', 0) or 0)) * 1024, file_size=byte2mb(int(float(edition.get('size', 0) or 0)) * 1024), identifier=song_id, duration_s=int(float(search_result.get('duration', 0) or 0)),
                    duration=seconds2hms(search_result.get('duration', 0) or 0), lyric=None, cover_url=safeextractfromdict(search_result, ['cover'], None), download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
                )
                if song_info.with_valid_download_url: break
            if song_info.with_valid_download_url: break
        # return
        return song_info
    '''_parsebytrack'''
    def _parsebytrack(self, search_results, song_infos: list = [], request_overrides: dict = None, progress: Progress = None):
        request_overrides = request_overrides or {}
        for search_result in search_results['data']['data']:
            if (not isinstance(search_result, dict)) or ('id' not in search_result) or (search_result.get('type') not in {'program'}): continue
            song_info = SongInfo(source=self.source)
            for parser in [self._parsewithofficialapiv1]:
                try: song_info = parser(search_result=search_result, request_overrides=request_overrides)
                except: continue
                if song_info.with_valid_download_url: break
            if not song_info.with_valid_download_url: continue
            song_infos.append(song_info)
            if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
        return song_infos
    '''_parsebyalbum'''
    def _parsebyalbum(self, search_results, song_infos: list = [], request_overrides: dict = None, progress: Progress = None):
        request_overrides = request_overrides or {}
        for search_result in search_results['data']['data']:
            if (not isinstance(search_result, dict)) or ('id' not in search_result) or (search_result.get('type') not in {'channel_ondemand'}): continue
            try: search_result['channel_info'] = self._fetchchannelinfo(search_result['id'], request_overrides)
            except Exception: pass
            download_results, page_size, tracks, unique_track_ids = [], 100, [], set()
            song_info = SongInfo(
                raw_data={'search': search_result, 'download': download_results, 'lyric': {}}, source=self.source, song_name=legalizestring(search_result.get('title')), singers=legalizestring(safeextractfromdict(search_result, ['podcaster', 'name'], None)), 
                album=f"{safeextractfromdict(search_result, ['channel_info', 'data', 'program_count'], 0) or 0} Episodes", ext=None, file_size=None, identifier=search_result['id'], duration='-:-:-', lyric=None, cover_url=search_result.get('cover', None),
                download_url=None, download_url_status={}, episodes=[],
            )
            num_pages = math.ceil(int(safeextractfromdict(search_result, ['channel_info', 'data', 'program_count'], 0) or 0) / page_size)
            download_album_pid = progress.add_task(f"{self.source}._parsebyalbum >>> (0/{num_pages}) pages downloaded in album {search_result['id']}", total=num_pages)
            for page_num_idx, page_num in enumerate(range(1, num_pages + 1)):
                if page_num_idx > 0:
                    progress.advance(download_album_pid, 1)
                    progress.update(download_album_pid, description=f"{self.source}._parsebyalbum >>> ({page_num_idx}/{num_pages}) pages downloaded in album {search_result['id']}")
                try: download_results.append(self._listpageprograms(search_result['id'], page=page_num, page_size=page_size, request_overrides=request_overrides))
                except: continue
            progress.advance(download_album_pid, 1)
            progress.update(download_album_pid, description=f"{self.source}._parsebyalbum >>> ({page_num_idx+1}/{num_pages}) pages downloaded in album {search_result['id']}")
            for download_result in download_results:
                for track in (safeextractfromdict(download_result, ['data', 'programs'], []) or []):
                    if not isinstance(track, dict) or not track.get('id'): continue
                    if track.get('id') in unique_track_ids: continue
                    unique_track_ids.add(track.get('id'))
                    tracks.append(track)
            download_album_pid = progress.add_task(f"{self.source}._parsebyalbum >>> (0/{len(tracks)}) episodes completed in album {search_result['id']}", total=len(tracks))
            for track_idx, track in enumerate(tracks):
                if track_idx > 0:
                    progress.advance(download_album_pid, 1)
                    progress.update(download_album_pid, description=f"{self.source}._parsebyalbum >>> ({track_idx}/{len(tracks)}) episodes completed in album {search_result['id']}")
                eps_info, track['channel_info'] = SongInfo(source=self.source), search_result.get('channel_info', {})
                track['url'] = f"qingtingfm://app.qingting.fm/playingview?type=ondemand&channel_id={search_result['id']}&program_id={track['id']}"
                for parser in [self._parsewithofficialapiv1]:
                    try: eps_info = parser(search_result=track, request_overrides=request_overrides)
                    except: continue
                    if eps_info.with_valid_download_url: break
                if not eps_info.with_valid_download_url: continue
                song_info.episodes.append(eps_info)
            progress.advance(download_album_pid, 1)
            progress.update(download_album_pid, description=f"{self.source}._parsebyalbum >>> ({track_idx+1}/{len(tracks)}) episodes completed in album {search_result['id']}")
            if not song_info.with_valid_download_url: continue
            try: song_info.duration_s = sum([eps.duration_s for eps in song_info.episodes]); song_info.duration = seconds2hms(song_info.duration_s)
            except Exception: pass
            try: song_info.file_size_bytes = sum([eps.file_size_bytes for eps in song_info.episodes]); song_info.file_size = byte2mb(song_info.file_size_bytes)
            except Exception: pass
            song_infos.append(song_info)
            if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
        return song_infos
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        # successful
        try:
            # --search results
            resp = self.get(search_url, **request_overrides)
            resp.raise_for_status()
            search_results = resp2json(resp)
            # --parse based on search type
            search_type = parse_qs(urlparse(search_url).query, keep_blank_values=True).get('include')[0]
            parsers = {'channel_ondemand': self._parsebyalbum, 'program_ondemand': self._parsebytrack}
            parsers[search_type](search_results, song_infos=song_infos, request_overrides=request_overrides, progress=progress)
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos


'''
Function:
    Implementation of KKWSMusicClient: https://www.kkws.cc/
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import re
from bs4 import BeautifulSoup
from urllib.parse import urljoin
from .base import BaseMusicClient
from rich.progress import Progress
from ..utils import legalizestring, usesearchheaderscookies, seconds2hms, searchdictbykey, safeextractfromdict, extractdurationsecondsfromlrc, cleanlrc, SongInfo, QuarkParser


'''KKWSMusicClient'''
class KKWSMusicClient(BaseMusicClient):
    source = 'KKWSMusicClient'
    MUSIC_QUALITY_RANK = {"DSD": 100, "DSF": 100, "DFF": 100, "WAV": 95, "AIFF": 95, "FLAC": 90, "ALAC": 90, "APE": 88, "WV": 88, "OPUS": 70, "AAC": 65, "M4A": 65, "OGG": 60, "VORBIS": 60, "MP3": 50, "WMA": 45}
    def __init__(self, **kwargs):
        super(KKWSMusicClient, self).__init__(**kwargs)
        assert self.quark_parser_config.get('cookies'), f'{self.source}.__init__ >>> "quark_parser_config" is not configured, so the songs cannot be downloaded.'
        self.default_search_headers = {
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/142.0.0.0 Safari/537.36",
        }
        self.default_download_headers = {
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/142.0.0.0 Safari/537.36",
        }
        self.default_headers = self.default_search_headers
        self._initsession()
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        # construct search urls
        self.search_size_per_page = min(self.search_size_per_source, 15)
        search_urls, page_size, count = [], self.search_size_per_page, 0
        while self.search_size_per_source > count:
            search_urls.append(f'https://www.kkws.cc/search.html?key={keyword}&page={int(count // page_size) + 1}')
            count += page_size
        # return
        return search_urls
    '''_parsesearchresultsfromhtml'''
    def _parsesearchresultsfromhtml(self, html_text: str):
        soup = BeautifulSoup(html_text, "lxml")
        search_results, base_url = [], 'https://www.kkws.cc/'
        for li in soup.select("ul.listbox > li"):
            a = li.select_one("h2 a[href]")
            if not a: continue
            href = urljoin(base_url, a["href"])
            title_attr = (a.get("title") or "").strip()
            full_text = a.get_text(" ", strip=True)
            name = title_attr.replace("免费下载", "").strip() if title_attr else full_text
            name = re.sub(r"\s*\[[^\]]+\]\s*", " ", name).strip()
            name = re.sub(r"\s*-\s*\d+(\.\d+)?[KMG]?\s*$", "", name).strip()
            m_fmt = re.search(r"\[([^\]]+)\]", full_text)
            file_format = m_fmt.group(1).strip() if m_fmt else ""
            m_size = re.search(r"-\s*([0-9.]+\s*[KMG]?)", full_text, re.IGNORECASE)
            size = (m_size.group(1).replace(" ", "") if m_size else "").strip()
            ems = li.select("small em")
            share_time, singer = "", ""
            if len(ems) >= 1: share_time = ems[0].get_text(strip=True).replace("分享时间：", "").strip()
            if len(ems) >= 2: singer = ems[-1].get_text(strip=True).replace("演唱：", "").strip()
            m_id = re.search(r"/detail/(\d+)\.html", href)
            item_id = m_id.group(1) if m_id else ""
            search_results.append({"id": item_id, "name": name, "format": file_format, "size": size, "share_time": share_time, "singer": singer, "detail_url": href})
        return search_results
    '''_extractlyricsandquark'''
    def _extractlyricsandquark(self, html_text: str):
        soup = BeautifulSoup(html_text, "lxml")
        tb = soup.select_one("#textbox")
        to_mmss_func = lambda t: (lambda s: f"{s//60:02d}:{s%60:02d}")(int(float(t.split(":",1)[0])*60+float(t.split(":",1)[1])) if ":" in t else int(float(t)))
        lyrics = "" if not tb else "\n".join((f"[{to_mmss_func(m.group(1))}] {m.group(2).strip()}" if (m:=re.match(r"^\[(\d+(?:\.\d+)?|\d{1,2}:\d{2}(?:\.\d+)?)\]\s*(.*)$", line)) else f"{line}") for line in (l.strip() for l in tb.get_text("\n").splitlines()) if line)
        url_map, rank = {}, KKWSMusicClient.MUSIC_QUALITY_RANK
        for a in soup.select("div.downbox a[onclick]"):
            onclick = (a.get("onclick") or "").strip()
            if not onclick: continue
            args = re.findall(r"'([^']*)'", onclick)
            name = fmt = url = None
            if onclick.startswith("openModel") and len(args) >= 4: name_fmt, url, fmt = args[1], args[2], (args[3] or None)
            elif onclick.startswith("mbgotourl") and len(args) >= 3: name_fmt, url, fmt = args[1], args[2], None
            else: continue
            if "|" in name_fmt: name, fmt2 = map(str.strip, name_fmt.split("|", 1)); fmt = fmt or fmt2
            else: name = name_fmt.strip()
            if not (url and "pan.quark.cn" in url): continue
            e = url_map.setdefault(url, {"url": url, "formats": set(), "names": set()})
            if fmt:
                for k in rank: fmt = k if k.lower() in fmt.lower() else fmt
                e["formats"].add(fmt)
            if name: e["names"].add(name)
        quark_links = sorted(({"url": e["url"], "formats": sorted(e["formats"]), "names": sorted(e["names"])} for e in url_map.values()), key=lambda x: rank.get(x["formats"][0] if x["formats"] else "UNKNOWN", 0), reverse=True)
        quark_links = [q for q in quark_links if isinstance(q, dict) and q.get('url')]
        return {"lyrics": lyrics, "quark_links": quark_links}
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        # successful
        try:
            # --search results
            resp = self.get(search_url, **request_overrides)
            resp.raise_for_status()
            search_results = self._parsesearchresultsfromhtml(resp.text)
            for search_result in search_results:
                # --download results
                if not isinstance(search_result, dict) or ('detail_url' not in search_result) or ('id' not in search_result): continue
                song_info = SongInfo(source=self.source)
                try: resp = self.get(search_result['detail_url'], **request_overrides); resp.raise_for_status(); download_result = self._extractlyricsandquark(resp.text)
                except: continue
                for quark_info in download_result['quark_links']:
                    quark_download_url = quark_info['url']
                    download_result['quark_parse_result'], download_url = QuarkParser.parsefromurl(quark_download_url, **self.quark_parser_config)
                    if not download_url or not str(download_url).startswith('http'): continue
                    duration = [int(float(d)) for d in searchdictbykey(download_result, 'duration') if int(float(d)) > 0]
                    duration_s = duration[0] if duration else 0
                    song_info = SongInfo(
                        raw_data={'search': search_result, 'download': download_result, 'lyric': {}}, source=self.source, song_name=legalizestring(safeextractfromdict(search_result, ['name'], None)),
                        singers=legalizestring(safeextractfromdict(search_result, ['singer'], None)), album='NULL', ext='mp3', file_size='NULL', identifier=search_result['id'], duration_s=duration_s,
                        duration=seconds2hms(duration_s), lyric=cleanlrc(safeextractfromdict(download_result, ['lyrics'], '')), cover_url=None, download_url=download_url,
                        download_url_status=self.quark_audio_link_tester.test(download_url, request_overrides), default_download_headers=self.quark_default_download_headers,
                    )
                    song_info.download_url_status['probe_status'] = self.quark_audio_link_tester.probe(song_info.download_url, request_overrides)
                    song_info.file_size = song_info.download_url_status['probe_status']['file_size']
                    song_info.ext = song_info.download_url_status['probe_status']['ext'] if (song_info.download_url_status['probe_status']['ext'] and song_info.download_url_status['probe_status']['ext'] not in ('NULL', )) else song_info.ext
                    if song_info.with_valid_download_url: break
                if not song_info.lyric or '歌词获取失败' in song_info.lyric: song_info.lyric = 'NULL'
                if not song_info.duration or song_info.duration == '-:-:-': song_info.duration = seconds2hms(extractdurationsecondsfromlrc(song_info.lyric))
                if not song_info.with_valid_download_url: continue
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos


'''initialize'''
from .data import SongInfo
from .hls import HLSDownloader
from .ip import RandomIPGenerator
from .quarkparser import QuarkParser
from .lanzouyparser import LanZouYParser
from .songinfoutils import SongInfoUtils
from .modulebuilder import BaseModuleBuilder
from .hosts import obtainhostname, hostmatchessuffix
from .importutils import optionalimport, optionalimportfrom
from .logger import LoggerHandle, colorize, printtable, printfullline, smarttrunctable, cursorpickintable
from .lyric import WhisperLRC, SodaTimedLyricsParser, extractdurationsecondsfromlrc, kuwolyricslisttolrc, cleanlrc
from .misc import (
    AudioLinkTester, legalizestring, touchdir, seconds2hms, byte2mb, cachecookies, resp2json, isvalidresp, safeextractfromdict, replacefile,
    usedownloadheaderscookies, useparseheaderscookies, usesearchheaderscookies, cookies2dict, cookies2string, estimatedurationwithfilesizebr,
    estimatedurationwithfilelink, searchdictbykey, shortenpathsinsonginfos
)


'''
Function:
    Implementation of KuwoMusicClient: http://www.kuwo.cn/
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import os
import re
import copy
import time
import random
import base64
import warnings
from .base import BaseMusicClient
from rich.progress import Progress
from pathvalidate import sanitize_filepath
from ..utils.hosts import KUWO_MUSIC_HOSTS
from ..utils.kuwoutils import KuwoMusicClientUtils
from urllib.parse import urlencode, urlparse, parse_qs
from rich.progress import Progress, TextColumn, BarColumn, TimeRemainingColumn, MofNCompleteColumn
from ..utils import touchdir, optionalimport, legalizestring, resp2json, seconds2hms, usesearchheaderscookies, safeextractfromdict, kuwolyricslisttolrc, useparseheaderscookies, obtainhostname, hostmatchessuffix, cleanlrc, SongInfo
warnings.filterwarnings('ignore')


'''KuwoMusicClient'''
class KuwoMusicClient(BaseMusicClient):
    source = 'KuwoMusicClient'
    MUSIC_QUALITIES = [(22000, 'flac'), (320, 'mp3')] # playable flac and mp3 formats
    ENC_MUSIC_QUALITIES = [(4000, '4000kflac'), (2000, '2000kflac'), (320, '320kmp3'), (192, '192kmp3'), (128, '128kmp3')] # encrypted mgg format
    def __init__(self, **kwargs):
        super(KuwoMusicClient, self).__init__(**kwargs)
        self.default_search_headers = {'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36'}
        self.default_download_headers = {'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36'}
        self.default_parse_headers = {'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36'}
        self.default_headers = self.default_search_headers
        self._initsession()
    '''_parsewithcggapi'''
    def _parsewithcggapi(self, search_result: dict, request_overrides: dict = None):
        # init
        curl_cffi = optionalimport('curl_cffi')
        request_overrides, song_id = request_overrides or {}, str(search_result.get('MUSICRID') or search_result.get('musicrid')).removeprefix('MUSIC_')
        MUSIC_QUALITIES = ["acc", "wma", "ogg", "standard", "exhigh", "ape", "lossless", "hires", "zp", "hifi", "sur", "jymaster"]
        # safe fetch filesize func
        safe_fetch_filesize_func = lambda meta: (lambda s: (lambda: float(s))() if s.replace('.', '', 1).isdigit() else 0)(str(meta.get('size', '0.00MB')).removesuffix('MB').strip()) if isinstance(meta, dict) else 0
        # parse
        for quality in MUSIC_QUALITIES[::-1][3:]:
            try: resp = curl_cffi.requests.get(f"https://kw-api.cenguigui.cn/?id={song_id}&type=song&level={quality}&format=json", timeout=10, impersonate="chrome131", verify=False, **request_overrides); resp.raise_for_status()
            except Exception: resp = self.get(f"https://kw-api.cenguigui.cn/?id={song_id}&type=song&level={quality}&format=json", timeout=10, **request_overrides); resp.raise_for_status()
            download_result = resp2json(resp=resp)
            if 'data' not in download_result or (safe_fetch_filesize_func(download_result['data']) < 1): continue
            download_url = safeextractfromdict(download_result, ['data', 'url'], '')
            if not download_url or not str(download_url).startswith('http'): continue
            lyric = cleanlrc(safeextractfromdict(download_result, ['data', 'lyric'], "NULL"))
            song_info = SongInfo(
                raw_data={'search': search_result, 'download': download_result, 'lyric': {}}, source=self.source, song_name=legalizestring(safeextractfromdict(download_result, ['data', 'name'], None)),
                singers=legalizestring(safeextractfromdict(download_result, ['data', 'artist'], None) or ""), album=legalizestring(safeextractfromdict(download_result, ['data', 'album'], None) or ""), 
                ext=download_url.split('?')[0].split('.')[-1], file_size=str(safeextractfromdict(download_result, ['data', 'size'], "")).removesuffix('MB').strip() + ' MB', identifier=str(song_id),
                duration_s=safeextractfromdict(download_result, ['data', 'duration'], 0), duration=seconds2hms(safeextractfromdict(download_result, ['data', 'duration'], 0)), lyric=lyric or "NULL",
                cover_url=safeextractfromdict(download_result, ['data', 'pic'], ""), download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
            )
            song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
            song_info.file_size = song_info.download_url_status['probe_status']['file_size']
            if song_info.with_valid_download_url: break
        return song_info
    '''_parsewithyyy001api'''
    def _parsewithyyy001api(self, search_result: dict, request_overrides: dict = None):
        # init
        decrypt_func = lambda t: base64.b64decode(str(t).encode('utf-8')).decode('utf-8')
        MUSIC_QUALITIES, REQUEST_KEYS = ["ff", "p", "h"], ['YzJmNjBlZDYtOTlmZC0xNjJlLWM0NzAtYjIxNDkwOGViNWI0YjYzYzFhN2E=', 'NTVjNTY3YzItNTJlNS1kMzdiLTE1N2MtMDE0MDIxNzEwYzc1NzY2OWNkYjc=', 'OTY4M2MwNzQtY2E3ZS01ZGYwLTUyZGEtMWEzNGZiNjVhOTZhZGU2NTczYjU=', 'OTdkZjQ0OTUtYzRjOS01MmFhLTNlODAtZjliZGFiODU1Y2UxZWIwN2JlZDk=']
        request_overrides, song_id = request_overrides or {}, str(search_result.get('MUSICRID') or search_result.get('musicrid')).removeprefix('MUSIC_')
        # parse
        for quality in MUSIC_QUALITIES:
            for _ in range(3):
                try: (resp := self.get(f"https://api.yyy001.com/api/kwmusic/?apikey={decrypt_func(random.choice(REQUEST_KEYS))}&action=music_url&music_id={song_id}&quality={quality}", timeout=10, **request_overrides)).raise_for_status(); assert resp.json()['code'] in {'200', 200}; break
                except Exception: time.sleep(1); continue
            download_result = resp2json(resp=resp)
            download_url = safeextractfromdict(download_result, ['data', 'url'], '')
            if not download_url or not str(download_url).startswith('http'): continue
            for _ in range(3):
                try: (resp := self.get(f"https://api.yyy001.com/api/kwmusic/?apikey={decrypt_func(random.choice(REQUEST_KEYS))}&action=music_info&music_id={song_id}", timeout=10, **request_overrides)).raise_for_status(); download_result['song_info'] = resp2json(resp=resp); assert resp.json()['code'] in {'200', 200}; break
                except Exception: time.sleep(1); continue
            for _ in range(3):
                try: (resp := self.get(f"https://api.yyy001.com/api/kwmusic/?apikey={decrypt_func(random.choice(REQUEST_KEYS))}&action=lyric&music_id={song_id}", timeout=10, **request_overrides)).raise_for_status(); lyric_result = resp2json(resp=resp); assert resp.json()['code'] in {'200', 200}; break
                except Exception: time.sleep(1); lyric_result = {}
            song_info = SongInfo(
                raw_data={'search': search_result, 'download': download_result, 'lyric': lyric_result}, source=self.source, song_name=legalizestring(safeextractfromdict(download_result, ['song_info', 'data', 'name'], None)),
                singers=legalizestring(safeextractfromdict(download_result, ['song_info', 'data', 'artist'], None) or ""), album=legalizestring(safeextractfromdict(download_result, ['song_info', 'data', 'album'], None)), 
                ext=download_url.split('?')[0].split('.')[-1], file_size_bytes='NULL', file_size='NULL', identifier=str(song_id), duration_s=safeextractfromdict(download_result, ['song_info', 'data', 'duration'], None),
                duration=seconds2hms(safeextractfromdict(download_result, ['song_info', 'data', 'duration'], None)), lyric=cleanlrc(kuwolyricslisttolrc(safeextractfromdict(lyric_result, ['data', 'lyric'], []) or [])) or 'NULL',
                cover_url=safeextractfromdict(download_result, ['song_info', 'data', 'images', 'pic'], ""), download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
            )
            song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
            song_info.file_size = song_info.download_url_status['probe_status']['file_size']
            if song_info.with_valid_download_url: break
        return song_info
    '''_parsewiththirdpartapis'''
    def _parsewiththirdpartapis(self, search_result: dict, request_overrides: dict = None):
        if self.default_cookies or request_overrides.get('cookies'): return SongInfo(source=self.source)
        for imp_func in [self._parsewithcggapi, self._parsewithyyy001api]:
            try:
                song_info_flac = imp_func(search_result, request_overrides)
                if song_info_flac.with_valid_download_url: break
            except:
                song_info_flac = SongInfo(source=self.source)
        return song_info_flac
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        # search rules
        default_rule = {
            "vipver": "1", "client": "kt", "ft": "music", "cluster": "0", "strategy": "2012", "encoding": "utf8", "rformat": "json", "mobi": "1", "issubtitle": "1", "show_copyright_off": "1", "pn": "0", "rn": "10", "all": keyword,
        }
        default_rule.update(rule)
        # construct search urls based on search rules
        base_url = 'http://www.kuwo.cn/search/searchMusicBykeyWord?'
        search_urls, page_size, count = [], self.search_size_per_page, 0
        while self.search_size_per_source > count:
            page_rule = copy.deepcopy(default_rule)
            page_rule['rn'] = page_size
            page_rule['pn'] = str(int(count // page_size))
            search_urls.append(base_url + urlencode(page_rule))
            count += page_size
        # return
        return search_urls
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        safe_fetch_filesize_func = lambda size: (lambda s: (lambda: float(s))() if s.replace('.', '', 1).isdigit() else 0)(size.removesuffix('MB').strip()) if isinstance(size, str) else 0
        # successful
        try:
            # --search results
            resp = self.get(search_url, **request_overrides)
            resp.raise_for_status()
            search_results = resp2json(resp)['abslist']
            for search_result in search_results:
                # --download results
                if not isinstance(search_result, dict) or ('MUSICRID' not in search_result): continue
                song_info = SongInfo(source=self.source)
                song_info_flac = self._parsewiththirdpartapis(search_result=search_result, request_overrides=request_overrides)
                song_info_flac = song_info
                for quality in KuwoMusicClient.MUSIC_QUALITIES:
                    if song_info_flac.with_valid_download_url and song_info_flac.ext in ('flac',): song_info = song_info_flac; break
                    query = f"user=0&corp=kuwo&source=kwplayer_ar_5.1.0.0_B_jiakong_vh.apk&p2p=1&type=convert_url2&sig=0&format={quality[1]}&rid={str(search_result['MUSICRID']).removeprefix('MUSIC_')}"
                    try: (resp := self.get(f"http://mobi.kuwo.cn/mobi.s?f=kuwo&q={KuwoMusicClientUtils.encryptquery(query)}", headers={"user-agent": "okhttp/3.10.0"}, **request_overrides)).raise_for_status(); download_result = resp.text
                    except Exception: continue
                    download_url = re.search(r'http[^\s$\"]+', download_result)
                    if not download_url: continue
                    download_url = download_url.group(0)
                    song_info = SongInfo(
                        raw_data={'search': search_result, 'download': download_result, 'lyric': {}}, source=self.source, song_name=legalizestring(safeextractfromdict(search_result, ['SONGNAME'], None)),
                        singers=legalizestring(safeextractfromdict(search_result, ['ARTIST'], None)), album=legalizestring(search_result.get('ALBUM', None)), ext=download_url.split('?')[0].split('.')[-1], 
                        file_size='NULL', identifier=str(search_result['MUSICRID']).removeprefix('MUSIC_'), duration_s=search_result.get('DURATION'), duration=seconds2hms(search_result.get('DURATION', 0)),
                        lyric='NULL', cover_url=safeextractfromdict(search_result, ['hts_MVPIC'], ""), download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
                    )
                    song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
                    song_info.file_size = song_info.download_url_status['probe_status']['file_size']
                    song_info.ext = song_info.download_url_status['probe_status']['ext'] if (song_info.download_url_status['probe_status']['ext'] and song_info.download_url_status['probe_status']['ext'] != 'NULL') else song_info.ext
                    if song_info_flac.with_valid_download_url and (safe_fetch_filesize_func(song_info.file_size) < safe_fetch_filesize_func(song_info_flac.file_size)): song_info = song_info_flac
                    if song_info.with_valid_download_url: break
                if not song_info.with_valid_download_url: song_info = song_info_flac
                if not song_info.with_valid_download_url: continue
                # --lyric results
                params = {'musicId': str(search_result['MUSICRID']).removeprefix('MUSIC_'), 'httpsStatus': '1'}
                try:
                    resp = self.get('http://m.kuwo.cn/newh5/singles/songinfoandlrc', params=params, **request_overrides)
                    resp.raise_for_status()
                    lyric_result: dict = resp2json(resp)
                    lyric = cleanlrc(kuwolyricslisttolrc(safeextractfromdict(lyric_result, ['data', 'lrclist'], [])))
                except:
                    lyric_result, lyric = {}, 'NULL'
                song_info.raw_data['lyric'] = lyric_result if lyric_result else song_info.raw_data['lyric']
                song_info.lyric = lyric if (lyric and (lyric not in {'NULL'})) else song_info.lyric
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos
    '''parseplaylist'''
    @useparseheaderscookies
    def parseplaylist(self, playlist_url: str, request_overrides: dict = None):
        request_overrides = request_overrides or {}
        playlist_url = self.session.head(playlist_url, allow_redirects=True, **request_overrides).url
        hostname = obtainhostname(url=playlist_url)
        if not hostname or not hostmatchessuffix(hostname, KUWO_MUSIC_HOSTS): return []
        try: playlist_id = parse_qs(urlparse(playlist_url).query, keep_blank_values=False).get('id')[0]; assert playlist_id
        except: playlist_id = urlparse(playlist_url).path.strip('/').split('/')[-1]
        page, tracks, song_infos = 1, [], []
        while True:
            try: resp = self.get(f"https://m.kuwo.cn/newh5app/wapi/api/www/playlist/playListInfo?pid={playlist_id}&pn={page}&rn=100", **request_overrides); resp.raise_for_status()
            except Exception: break
            playlist_results = resp2json(resp=resp)
            if (not safeextractfromdict(playlist_results, ['data', 'musicList'], [])) or (float(safeextractfromdict(playlist_results, ['data', 'total'], 0)) <= len(tracks)): break
            tracks.extend(safeextractfromdict(playlist_results, ['data', 'musicList'], []))
        tracks = list({d["musicrid"]: d for d in tracks}.values())
        with Progress(TextColumn("{task.description}"), BarColumn(bar_width=None), MofNCompleteColumn(), TimeRemainingColumn(), refresh_per_second=10) as main_process_context:
            main_progress_id = main_process_context.add_task(f"{len(tracks)} songs found in playlist {playlist_id} >>> completed (0/{len(tracks)})", total=len(tracks))
            for idx, track_info in enumerate(tracks):
                if idx > 0: main_process_context.advance(main_progress_id, 1)
                main_process_context.update(main_progress_id, description=f"{len(tracks)} songs found in playlist {playlist_id} >>> completed ({idx}/{len(tracks)})")
                for third_part_api in [self._parsewithcggapi, self._parsewithyyy001api]:
                    try:
                        song_info = third_part_api(track_info, request_overrides=request_overrides)
                        if song_info.with_valid_download_url: song_infos.append(song_info); break
                    except:
                        continue
            main_process_context.advance(main_progress_id, 1)
            main_process_context.update(main_progress_id, description=f"{len(tracks)} songs found in playlist {playlist_id} >>> completed ({idx+1}/{len(tracks)})")
        song_infos = self._removeduplicates(song_infos=song_infos)
        work_dir = self._constructuniqueworkdir(keyword=playlist_id)
        for song_info in song_infos:
            song_info.work_dir = work_dir; episodes = song_info.episodes if isinstance(song_info.episodes, list) else []
            for eps_info in episodes: eps_info.work_dir = sanitize_filepath(os.path.join(work_dir, song_info.song_name)); touchdir(work_dir)
        return song_infos


'''initialize'''
from .sources import MusicClientBuilder, BaseMusicClient, BuildMusicClient
from .utils import (
    BaseModuleBuilder, LoggerHandle, AudioLinkTester, WhisperLRC, QuarkParser, SongInfo, SongInfoUtils, RandomIPGenerator, SodaTimedLyricsParser, LanZouYParser,
    HLSDownloader, cachecookies, resp2json, isvalidresp, safeextractfromdict, replacefile, printfullline, smarttrunctable, usesearchheaderscookies, byte2mb, seconds2hms,
    usedownloadheaderscookies, useparseheaderscookies, cookies2dict, cookies2string, touchdir, estimatedurationwithfilesizebr, estimatedurationwithfilelink,
    extractdurationsecondsfromlrc, searchdictbykey, colorize, optionalimportfrom, legalizestring, kuwolyricslisttolrc, shortenpathsinsonginfos, cursorpickintable, 
    printtable, optionalimport, obtainhostname, hostmatchessuffix, cleanlrc
)


'''
Function:
    Implementation of SongInfoUtils
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
from __future__ import annotations
import os
import base64
import requests
from pathlib import Path
from mutagen import File
from .data import SongInfo
from tinytag import TinyTag
from .lyric import WhisperLRC
from mimetypes import guess_type
from .logger import LoggerHandle
from mutagen.flac import Picture
from mutagen.mp4 import MP4Cover
from .misc import seconds2hms, byte2mb
from mutagen.id3 import ID3, APIC, USLT
from .importutils import optionalimportfrom


'''SongInfoUtils'''
class SongInfoUtils:
    '''fillsongtechinfo'''
    @staticmethod
    def fillsongtechinfo(song_info: SongInfo, logger_handle: LoggerHandle, disable_print: bool, auto_write_tags_to_downloaded_audio: bool = True) -> SongInfo:
        path = Path(song_info.save_path)
        # correct file size
        size = path.stat().st_size
        song_info.file_size_bytes = size
        song_info.file_size = byte2mb(size=size)
        # tinytag parse
        try: tag = TinyTag.get(str(path))
        except Exception as err: logger_handle.warning(f'SongInfoUtils.fillsongtechinfo >>> {str(path)} (Err: {err})', disable_print=disable_print); return song_info
        if tag.duration: song_info.duration_s = int(round(tag.duration)); song_info.duration = seconds2hms(tag.duration)
        if tag.bitrate: song_info.bitrate = int(round(tag.bitrate))
        if tag.samplerate: song_info.samplerate = int(tag.samplerate)
        if tag.channels: song_info.channels = int(tag.channels)
        if getattr(tag, "codec", None): song_info.codec = tag.codec
        elif getattr(tag, "extra", None) and isinstance(tag.extra, dict): song_info.codec = tag.extra.get("codec") or tag.extra.get("mime-type")
        # lyric
        if os.environ.get('ENABLE_WHISPERLRC', 'False').lower() == 'true' and ((not song_info.lyric) or (song_info.lyric == 'NULL')):
            lyric_result = WhisperLRC(model_size_or_path='small').fromfilepath(str(path))
            lyric = lyric_result['lyric']
            song_info.lyric = lyric
            song_info.raw_data['lyric'] = lyric_result
        # write tags to audio file
        if auto_write_tags_to_downloaded_audio:
            try: SongInfoUtils.writetagstoaudio(song_info, overwrite=False)
            except: pass
        # return
        return song_info
    '''writetagstoaudio'''
    @staticmethod
    def writetagstoaudio(song_info: SongInfo, overwrite: bool = False):
        audio_path = Path(song_info.save_path)
        easy = File(audio_path, easy=True)
        if easy is None: raise ValueError(f"Unsupported/unreadable audio file: {audio_path}")
        tags, changed = {"artist": song_info.singers, "album": song_info.album, "title": song_info.song_name}, False
        for k, v in (tags or {}).items(): (not v or v == 'NULL') or ((overwrite or not easy.get(k)) and (easy.__setitem__(k, v if isinstance(v, (list, tuple)) else str(v)) or (changed := True)))
        if changed: 
            try: easy.save(v2_version=3)
            except: easy.save()
        if song_info.cover_url and song_info.cover_url != 'NULL' and (str(song_info.cover_url).startswith('http') or os.path.exists(str(song_info.cover_url))):
            try: changed = SongInfoUtils.writecovertoaudio(audio_path, song_info.cover_url, overwrite=overwrite) or changed
            except: pass
        if song_info.lyric and song_info.lyric != 'NULL':
            try: changed = SongInfoUtils.writelyricstoaudio(audio_path, song_info.lyric, overwrite=overwrite) or changed
            except: pass
        return changed
    '''loadimagebytesandmime'''
    @staticmethod
    def loadimagebytesandmime(cover: str | Path, *, timeout: int = 15) -> tuple[bytes, str]:
        cover_str, headers = str(cover), {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36'}
        # local image path
        if not cover_str.startswith('http'):
            path = Path(cover)
            data = path.read_bytes()
            mime = (guess_type(str(path))[0] or "image/jpeg").split(";")[0]
            return data, mime
        # network image url
        resp = requests.get(cover_str, stream=True, timeout=timeout, headers=headers, allow_redirects=True)
        resp.raise_for_status()
        ctype = (resp.headers.get("Content-Type") or "").split(";")[0].strip().lower()
        mime, data = (ctype or (guess_type(cover_str)[0] or "image/jpeg")).split(";")[0], resp.content
        if not mime.startswith("image/"):
            sig = data[:12]
            if sig.startswith(b"\xFF\xD8\xFF"): mime = "image/jpeg"
            elif sig.startswith(b"\x89PNG\r\n\x1a\n"): mime = "image/png"
            elif sig[:4] == b"RIFF" and sig[8:12] == b"WEBP": mime = "image/webp"
            elif sig[:6] in (b"GIF87a", b"GIF89a"): mime = "image/gif"
            else: raise ValueError(f"URL did not return an image (Content-Type={ctype!r}).")
        return data, mime
    '''writecovertoaudio'''
    @staticmethod
    def writecovertoaudio(audio_path: Path, cover_path: str | Path, *, overwrite: bool = False):
        # init
        cover_bytes, mime = SongInfoUtils.loadimagebytesandmime(cover_path)
        audio = File(audio_path)
        if audio is None: raise ValueError(f"Unsupported/unreadable audio file: {audio_path}")
        cls, ext = audio.__class__.__name__, audio_path.suffix.lower()
        # mp3
        if ext == ".mp3":
            try: id3 = ID3(audio_path)
            except Exception: id3 = ID3()
            if not overwrite and any(k.startswith("APIC") for k in id3.keys()): return False
            if overwrite: id3.delall("APIC")
            id3.add(APIC(encoding=3, mime=mime, type=3, desc="Cover", data=cover_bytes))
            id3.save(audio_path, v2_version=3)
            return True
        # mp4
        if cls == "MP4":
            has_cover = bool(audio.tags) and ("covr" in audio.tags)
            if has_cover and not overwrite: return False
            imgfmt = MP4Cover.FORMAT_PNG if mime == "image/png" else MP4Cover.FORMAT_JPEG
            audio["covr"] = [MP4Cover(cover_bytes, imageformat=imgfmt)]
            audio.save()
            return True
        # flac
        if cls == "FLAC":
            has_cover = bool(getattr(audio, "pictures", []))
            if has_cover and not overwrite: return False
            pic = Picture()
            pic.type, pic.mime, pic.desc, pic.data = 3, mime, "Cover", cover_bytes
            if overwrite: audio.clear_pictures()
            audio.add_picture(pic)
            audio.save()
            return True
        # ogg
        if cls in {"OggVorbis", "OggOpus", "OggSpeex", "OggTheora"}:
            has_cover = bool(audio.tags) and ("METADATA_BLOCK_PICTURE" in audio.tags)
            if has_cover and not overwrite: return False
            pic = Picture()
            pic.type, pic.mime, pic.desc, pic.data = 3, mime, "Cover", cover_bytes
            audio["METADATA_BLOCK_PICTURE"] = [base64.b64encode(pic.write()).decode("ascii")]
            audio.save()
            return True
        # asf
        if cls == "ASF" or ext in {".wma", ".asf"}:
            has_cover = bool(audio.tags) and ("WM/Picture" in audio.tags)
            if has_cover and not overwrite: return False
            ASFPicture, ASFValue = optionalimportfrom('mutagen.asf', 'ASFPicture'), optionalimportfrom('mutagen.asf', 'ASFValue')
            if ASFPicture is not None:
                pic = ASFPicture()
                pic.type, pic.mime_type, pic.description, pic.data = 3, mime, "Cover", cover_bytes
                audio["WM/Picture"] = [pic]
                audio.save()
                return True
            if ASFValue is not None:
                audio["WM/Picture"] = [ASFValue(cover_bytes)]
                audio.save()
                return True
            raise NotImplementedError("ASF(WMA) cover embedding not available in your mutagen build (ASFPicture/ASFValue not exposed). Upgrade mutagen or skip WMA cover.")
        # not supported
        raise NotImplementedError(f"Cover embedding not supported for: {cls} ({ext})")
    '''writelyricstoaudio'''
    def writelyricstoaudio(audio_path: Path, lyrics: str, *, overwrite: bool) -> bool:
        # init
        audio = File(audio_path)
        if audio is None: raise ValueError(f"Unsupported/unreadable audio file: {audio_path}")
        cls, ext = audio.__class__.__name__, audio_path.suffix.lower()
        # mp3
        if ext == ".mp3":
            try: id3 = ID3(audio_path)
            except Exception: id3 = ID3()
            has_lyrics = any(k.startswith("USLT") for k in id3.keys())
            if has_lyrics and not overwrite: return False
            if overwrite: any(id3.__delitem__(k) for k in list(id3.keys()) if k.startswith("USLT"))
            id3.add(USLT(encoding=3, lang="eng", desc="Lyrics", text=lyrics))
            id3.save(audio_path)
            return True
        # m4a
        if cls == "MP4":
            has_lyrics = bool(audio.tags) and ("\xa9lyr" in audio.tags)
            if has_lyrics and not overwrite: return False
            audio["\xa9lyr"] = [lyrics]
            audio.save()
            return True
        # flac / ogg
        if cls in {"FLAC", "OggVorbis", "OggOpus", "OggSpeex", "OggTheora"}:
            has_lyrics = bool(audio.tags) and ("LYRICS" in audio.tags)
            if has_lyrics and not overwrite: return False
            audio["LYRICS"] = [lyrics]
            audio.save()
            return True
        # asf
        if cls == "ASF" or ext in {".wma", ".asf"}:
            has_lyrics = bool(audio.tags) and ("WM/Lyrics" in audio.tags)
            if has_lyrics and not overwrite: return False
            audio["WM/Lyrics"] = [lyrics]
            audio.save()
            return True
        # not supported
        raise NotImplementedError(f"Lyrics embedding not supported for: {cls} ({ext})")


'''
Function:
    Implementation of HLSDownloader
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import os
import re
import copy
import time
import math
import m3u8
import base64
import shutil
import hashlib
import requests
import threading
import concurrent.futures as cf
from pathlib import Path
from .misc import touchdir
from .logger import LoggerHandle
from urllib.parse import urljoin
from dataclasses import dataclass
from rich.progress import Progress
from typing import Optional, Dict, Any, Tuple, List, Union, Callable
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes


'''SegmentJob'''
@dataclass(frozen=True)
class SegmentJob:
    index: int
    uri: str
    byterange: Optional[str]
    key_method: Optional[str]
    key_uri: Optional[str]
    key_iv: Optional[str]
    keyformat: Optional[str]
    media_sequence: int
    map_uri: Optional[str]
    map_byterange: Optional[str]


'''HLSDownloader'''
class HLSDownloader:
    def __init__(self, output_dir: str = "downloads", proxies: Optional[Dict[str, str]] = None, headers: Optional[Dict[str, str]] = None, cookies: Optional[Dict[str, str]] = None, timeout: Tuple[float, float] = (10.0, 30.0), logger_handle: LoggerHandle = None,
                 verify_tls: bool = True, concurrency: int = 16, max_retries: int = 8, backoff_base: float = 0.6, backoff_cap: float = 10.0, chunk_size: int = 1024 * 256, strict_key_length: bool = False, disable_print: bool = False, request_overrides: dict = None):
        # work dir
        self.output_dir = output_dir
        touchdir(self.output_dir)
        # logger
        self.logger_handle = logger_handle
        self.disable_print = disable_print
        # http requests
        self.proxies = proxies or {}
        self.headers = headers or {}
        self.cookies = cookies or {}
        self.timeout = timeout
        self.verify_tls = verify_tls
        self.chunk_size = int(chunk_size)
        self.backoff_cap = float(backoff_cap)
        self.backoff_base = float(backoff_base)
        self.concurrency = max(1, int(concurrency))
        self.max_retries = max(1, int(max_retries))
        self.strict_key_length = bool(strict_key_length)
        self.request_overrides = request_overrides or {}
        # threading
        self._tls = threading.local()
        self._key_cache: Dict[str, bytes] = {}
        self._key_cache_lock = threading.Lock()
    '''download'''
    def download(self, m3u8_url: str, output_path: str, quality: Union[str, int, Callable[[List[Dict[str, Any]]], int]] = "best", keep_segments: bool = False, temp_subdir: Optional[str] = None, progress: Progress = None, progress_id: int = 0) -> str:
        master_or_media = self._loadm3u8(m3u8_url)
        if master_or_media.is_variant:
            variant_url = self._selectvariant(master_or_media, quality)
            self.logger_handle.info(f"Selected variant: {variant_url}", disable_print=self.disable_print)
            playlist = self._loadm3u8(variant_url)
        else:
            playlist = master_or_media
        jobs, global_init_map = self._buildjobs(playlist)
        temp_folder, global_init_path = os.path.join(self.output_dir, temp_subdir or f".hls_tmp_{self._safenamefromurl(m3u8_url)}"), None
        touchdir(temp_folder)
        if global_init_map:
            global_init_path = os.path.join(temp_folder, "_global_init.bin")
            if not self._fileok(global_init_path): self._atomicwrite(global_init_path, self._fetchbytes(global_init_map["uri"], global_init_map.get("byterange")))
        seg_paths = self._downloadallsegments(jobs, temp_folder, progress=progress, progress_id=progress_id)
        touchdir(os.path.dirname(os.path.abspath(output_path)) or ".")
        self._mergefiles(global_init_path, seg_paths, output_path)
        if not keep_segments: shutil.rmtree(temp_folder, ignore_errors=True)
        return output_path
    '''_getsession'''
    def _getsession(self) -> requests.Session:
        sess = getattr(self._tls, "session", None)
        if sess is None:
            sess = requests.Session()
            sess.headers.update(self.headers)
            if self.cookies: sess.cookies.update(self.cookies)
            self._tls.session = sess
        return sess
    '''_request'''
    def _request(self, url: str, method: str = "GET", headers: Optional[Dict[str, str]] = None, stream: bool = False, **kwargs) -> requests.Response:
        kwargs.update(copy.deepcopy(self.request_overrides))
        sess, last_exc = self._getsession(), None
        hdrs = dict(self.headers)
        if headers: hdrs.update(headers)
        for attempt in range(1, self.max_retries + 1):
            try:
                resp = sess.request(method=method, url=url, headers=hdrs, proxies=self.proxies, timeout=self.timeout, verify=self.verify_tls, stream=stream, **kwargs)
                if resp.status_code in (429, 500, 502, 503, 504): resp.close(); raise requests.HTTPError(f"HTTP {resp.status_code} for {url}")
                resp.raise_for_status()
                return resp
            except Exception as e:
                last_exc = e
                t = min(self.backoff_cap, self.backoff_base * (2 ** (attempt - 1)))
                t = t + (0.1 * t * (0.5 - (time.time() % 1)))
                time.sleep(max(0.0, t))
        raise RuntimeError(f"Request failed after retries: {url}\nLast error: {last_exc}")
    '''_gettext'''
    def _gettext(self, url: str) -> str:
        resp = self._request(url, stream=False)
        return resp.text
    '''_getbytes'''
    def _getbytes(self, url: str, headers: Optional[Dict[str, str]] = None) -> bytes:
        resp = self._request(url, headers=headers, stream=True)
        chunks = []
        for c in resp.iter_content(chunk_size=self.chunk_size):
            if c: chunks.append(c)
        resp.close()
        return b"".join(chunks)
    '''_fetchbytes'''
    def _fetchbytes(self, url: str, byterange: Optional[str]) -> bytes:
        headers = {}
        if byterange:
            length, offset = self._parsebyterange(byterange)
            headers["Range"] = f"bytes={offset}-{offset + length - 1}"
        return self._getbytes(url, headers=headers)
    '''_loadm3u8'''
    def _loadm3u8(self, url: str) -> m3u8.M3U8:
        text = self._gettext(url)
        return m3u8.loads(text, uri=url)
    '''_selectvariant'''
    def _selectvariant(self, master: m3u8.M3U8, quality: Union[str, int, Callable[[List[Dict[str, Any]]], int]]) -> str:
        variants, bw_func = [], lambda v: int(v.get("average_bandwidth") or v.get("bandwidth") or 0)
        for i, p in enumerate(master.playlists or []):
            si = getattr(p, "stream_info", None)
            variants.append({
                "index": i, "absolute_uri": getattr(p, "absolute_uri", None) or urljoin(master.base_uri or master.uri, p.uri), "uri": p.uri, "bandwidth": getattr(si, "bandwidth", None) if si else None,
                "average_bandwidth": getattr(si, "average_bandwidth", None) if si else None, "resolution": getattr(si, "resolution", None) if si else None, "codecs": getattr(si, "codecs", None) if si else None,
                "frame_rate": getattr(si, "frame_rate", None) if si else None,
            })
        if not variants: raise ValueError("Master playlist has no variants.")
        if callable(quality): idx = int(quality(variants)); idx = max(0, min(idx, len(variants) - 1)); return variants[idx]["absolute_uri"]
        if isinstance(quality, str):
            q = quality.lower().strip()
            if q == "best":
                chosen = max(variants, key=bw_func)
            elif q == "lowest":
                chosen = min(variants, key=bw_func)
            else:
                m = re.search(r"(\d+)", q)
                if m: target = int(m.group(1)); chosen = min(variants, key=lambda v: abs(bw_func(v) - target))
                else: chosen = max(variants, key=bw_func)
        else:
            target = int(quality)
            chosen = min(variants, key=lambda v: abs(bw_func(v) - target))
        return chosen["absolute_uri"]
    '''_buildjobs'''
    def _buildjobs(self, playlist: m3u8.M3U8) -> Tuple[List[SegmentJob], Optional[Dict[str, Any]]]:
        media_seq = int(getattr(playlist, "media_sequence", 0) or 0)
        global_init, seg_map = None, getattr(playlist, "segment_map", None)
        if seg_map:
            try: sm0 = seg_map[0]; global_init = {"uri": getattr(sm0, "absolute_uri", None) or urljoin(playlist.base_uri, sm0.uri), "byterange": getattr(sm0, "byterange", None)}
            except Exception: global_init = None
        jobs: List[SegmentJob] = []
        session_keys = getattr(playlist, "session_keys", None) or []
        fallback_session_key, last_key_obj = session_keys[-1] if session_keys else None, None
        for i, seg in enumerate(playlist.segments or []):
            seg_uri, key_obj = getattr(seg, "absolute_uri", None) or urljoin(playlist.base_uri, seg.uri), getattr(seg, "key", None) or last_key_obj or fallback_session_key
            if getattr(seg, "key", None) is not None: last_key_obj = getattr(seg, "key", None)
            key_method, key_uri, key_iv, keyformat = (getattr(key_obj, k, None) for k in ("method", "uri", "iv", "keyformat")) if key_obj else (None, None, None, None)
            key_uri_abs = (key_uri if key_uri and (key_uri.startswith("data:") or key_uri.startswith("skd://")) else (urljoin(playlist.base_uri, key_uri) if key_uri else None))
            init_section = getattr(seg, "init_section", None)
            map_uri, map_byterange = ((getattr(init_section, "absolute_uri", None) or (urljoin(playlist.base_uri, getattr(init_section, "uri", "")) if getattr(init_section, "uri", None) else None)), getattr(init_section, "byterange", None)) if init_section is not None else (None, None)
            jobs.append(SegmentJob(index=i, uri=seg_uri, byterange=getattr(seg, "byterange", None), key_method=key_method, key_uri=key_uri_abs, key_iv=key_iv, keyformat=keyformat, media_sequence=media_seq, map_uri=map_uri, map_byterange=map_byterange))
        return jobs, global_init
    '''_downloadallsegments'''
    def _downloadallsegments(self, jobs: List[SegmentJob], temp_folder: str, progress: Progress, progress_id: int) -> List[str]:
        progress.update(progress_id, description=f"HLSDownloader._downloadallsegments >>> completed (0/{len(jobs)})", total=len(jobs), kind='hls')
        byterange_cursor: Dict[str, int] = {}; seg_paths: List[Optional[str]] = [None] * len(jobs)
        init_cache: Dict[str, str] = {}; init_inflight: Dict[str, threading.Event] = {}; init_cache_lock = threading.Lock()
        def _ensureinitsection(map_uri: str, map_byterange: Optional[str]) -> bytes:
            key = f"{map_uri}|{map_byterange or ''}"
            with init_cache_lock:
                cached = init_cache.get(key)
                if cached and self._fileok(cached): return Path(cached).read_bytes()
                leader = (evt := init_inflight.get(key)) is None; evt = init_inflight[key] = threading.Event() if leader else evt
            if not leader:
                evt.wait()
                with init_cache_lock: cached = init_cache.get(key)
                return Path(cached).read_bytes() if cached and self._fileok(cached) else (_ for _ in ()).throw(RuntimeError(f"init_section download failed: {key}"))
            try:
                data = self._fetchbytes(map_uri, map_byterange)
                path = os.path.join(temp_folder, f"_initsec_{abs(hash(key)) & 0xffffffff:08x}.bin")
                self._atomicwrite(path, data)
                with init_cache_lock: init_cache[key] = path
                return data
            finally:
                with init_cache_lock: (evt := init_inflight.pop(key, None)) and evt.set()
        def _worker(job: SegmentJob) -> Tuple[int, str]:
            seg_path = os.path.join(temp_folder, f"seg_{job.index:06d}.bin")
            if self._fileok(seg_path): return job.index, seg_path
            prepend = _ensureinitsection(job.map_uri, job.map_byterange) if job.map_uri else b""
            eff_byterange = self._normalizebyterange(job.uri, job.byterange, byterange_cursor) if job.byterange else job.byterange
            data = self._fetchandmaybedecrypt(job, eff_byterange)
            self._atomicwrite(seg_path, prepend + data)
            return job.index, seg_path
        exceptions: List[Exception] = []
        with cf.ThreadPoolExecutor(max_workers=self.concurrency) as ex:
            futures = [ex.submit(_worker, j) for j in jobs]
            for fut in cf.as_completed(futures):
                try:
                    idx, path = fut.result()
                    seg_paths[idx] = path
                except Exception as e:
                    exceptions.append(e)
                finally:
                    progress.advance(progress_id, 1)
                    num_downloaded_segs = int(progress.tasks[progress_id].completed)
                    progress.update(progress_id, description=f"HLSDownloader._downloadallsegments >>> completed ({num_downloaded_segs}/{len(jobs)})")
        if exceptions: raise exceptions[0]
        return [p for p in seg_paths if p is not None]
    '''_fetchandmaybedecrypt'''
    def _fetchandmaybedecrypt(self, job: SegmentJob, eff_byterange: Optional[str]) -> bytes:
        method_raw, keyformat = (job.key_method or "").strip(), (job.keyformat or "").strip().lower()
        if not method_raw or method_raw.upper() == "NONE": return self._fetchbytes(job.uri, eff_byterange)
        if keyformat and keyformat not in ("identity",): raise NotImplementedError(f"Unsupported KEYFORMAT={job.keyformat} (likely DRM).")
        method = method_raw.upper().replace("_", "-")
        dec_mode = self._classifyencryptionmethod(method)
        if dec_mode in ("DRM", "UNSUPPORTED"): raise NotImplementedError(f"Unsupported encryption method: {method_raw}")
        if not job.key_uri: raise RuntimeError(f"Encrypted segment missing key URI at seg {job.index}")
        key, base_iv = self._prepareaeskey(method, self._getkeybytes(job.key_uri)), self._deriveiv(job.key_iv, job.media_sequence + job.index)
        if not eff_byterange: ciphertext = self._fetchbytes(job.uri, None); return self._decryptwhole(ciphertext, dec_mode, key, base_iv)
        length, offset = self._parsebyterange(eff_byterange)
        block, end = 16, offset + length
        aligned_start, aligned_end = (offset // block) * block, int(math.ceil(end / block) * block)
        if dec_mode == "CBC":
            fetch_start, drop = ((aligned_start - block, offset - aligned_start + block) if aligned_start > 0 else (aligned_start, offset - aligned_start)); fetch_len = aligned_end - fetch_start; fetch_range = f"{fetch_len}@{fetch_start}"
            ciphertext = self._fetchbytes(job.uri, fetch_range)
            iv = (b"\x00" * 16) if fetch_start > 0 else base_iv
            plaintext = self._aescbcdecrypt(ciphertext, key, iv)
            return plaintext[drop: drop+length]
        else:
            fetch_start, drop, fetch_len, fetch_range = aligned_start, offset - aligned_start, aligned_end - aligned_start, f"{aligned_end - aligned_start}@{aligned_start}"
            ciphertext = self._fetchbytes(job.uri, fetch_range)
            block_index = fetch_start // block
            iv_int = int.from_bytes(base_iv, "big")
            adj_iv = ((iv_int + block_index) % (1 << 128)).to_bytes(16, "big")
            plaintext = self._aesctrcrypt(ciphertext, key, adj_iv)
            return plaintext[drop: drop+length]
    '''_decryptwhole'''
    def _decryptwhole(self, ciphertext: bytes, dec_mode: str, key: bytes, iv: bytes) -> bytes:
        if dec_mode == "CBC": return self._aescbcdecrypt(ciphertext, key, iv)
        if dec_mode == "CTR": return self._aesctrcrypt(ciphertext, key, iv)
        raise NotImplementedError(f"decrypt mode {dec_mode} not supported")
    '''_classifyencryptionmethod'''
    def _classifyencryptionmethod(self, method: str) -> str:
        m = method.strip().upper()
        if m in ("AES-128", "AES-128-CBC", "AES-CBC", "CBC"): return "CBC"
        if m in ("AES-CTR", "AES-128-CTR", "AES-192-CTR", "AES-256-CTR"): return "CTR"
        if m.startswith("SAMPLE-AES") or "SKD" in m: return "DRM"
        return "UNSUPPORTED"
    '''_getkeybytes'''
    def _getkeybytes(self, key_uri: str) -> bytes:
        if key_uri.startswith("data:"):
            if "base64," in key_uri: b64 = key_uri.split("base64,", 1)[1]; return base64.b64decode(b64)
            if "," in key_uri: raw = key_uri.split(",", 1)[1]; return raw.encode("utf-8", errors="ignore")
            raise ValueError("Unsupported data: key URI")
        if key_uri.startswith("skd://"): raise NotImplementedError("skd:// indicates DRM (FairPlay). Not supported.")
        with self._key_cache_lock:
            if key_uri in self._key_cache: return self._key_cache[key_uri]
        b = self._getbytes(key_uri)
        with self._key_cache_lock: self._key_cache[key_uri] = b
        return b
    '''_decodekeyguess'''
    def _decodekeyguess(self, key_bytes: bytes) -> bytes:
        b = key_bytes.strip()
        if b"\x00" in b: return b
        b2 = b
        if b2.lower().startswith(b"0x"): b2 = b2[2:]
        if re.fullmatch(rb"[0-9a-fA-F]+", b2) and len(b2) in (32, 48, 64):
            try: return bytes.fromhex(b2.decode("ascii"))
            except Exception: pass
        if re.fullmatch(rb"[A-Za-z0-9+/=\r\n]+", b) and (len(b) % 4 == 0):
            try:
                dec = base64.b64decode(b, validate=False)
                if len(dec) in (16, 24, 32): return dec
            except Exception:
                pass
        return b
    '''_expectedkeylen'''
    def _expectedkeylen(self, method: str) -> int:
        m = method.upper()
        if "256" in m: return 32
        if "192" in m: return 24
        return 16
    '''_prepareaeskey'''
    def _prepareaeskey(self, method: str, key_bytes: bytes) -> bytes:
        k = self._decodekeyguess(key_bytes)
        want = self._expectedkeylen(method)
        if len(k) == want: return k
        if self.strict_key_length: raise ValueError(f"Bad key length for {method}: got {len(k)} bytes, expected {want}")
        self.logger_handle.warning(f"Key length mismatch for {method}: got {len(k)}, expected {want}. Best-effort fix.", disable_print=self.disable_print)
        if len(k) > want: return k[:want]
        return (k + b"\x00" * want)[:want]
    '''_deriveiv'''
    def _deriveiv(self, iv_str: Optional[str], seq_num: int) -> bytes:
        if not iv_str: return seq_num.to_bytes(16, byteorder="big", signed=False)
        s = str(iv_str).strip().lower()
        if s.startswith("0x"): s = s[2:]
        try: iv = bytes.fromhex(s)
        except Exception: iv = s.encode("utf-8", errors="ignore")
        if len(iv) < 16: iv = (b"\x00" * (16 - len(iv))) + iv
        if len(iv) > 16: iv = iv[-16:]
        return iv
    '''_aescbcdecrypt'''
    def _aescbcdecrypt(self, ciphertext: bytes, key: bytes, iv: bytes) -> bytes:
        if len(ciphertext) % 16 != 0: raise ValueError(f"CBC ciphertext length not multiple of 16: {len(ciphertext)} bytes")
        cipher = Cipher(algorithms.AES(key), modes.CBC(iv))
        dec = cipher.decryptor()
        return dec.update(ciphertext) + dec.finalize()
    '''_aesctrcrypt'''
    def _aesctrcrypt(self, data: bytes, key: bytes, iv: bytes) -> bytes:
        cipher = Cipher(algorithms.AES(key), modes.CTR(iv))
        dec = cipher.decryptor()
        return dec.update(data) + dec.finalize()
    '''_parsebyterange'''
    def _parsebyterange(self, s: str) -> Tuple[int, int]:
        s = s.strip()
        if "@" in s: a, b = s.split("@", 1); return int(a), int(b)
        raise ValueError(f"BYTERANGE missing offset: {s}")
    '''_normalizebyterange'''
    def _normalizebyterange(self, uri: str, byterange: str, cursor: Dict[str, int]) -> str:
        s = byterange.strip()
        if "@" in s: length, offset = s.split("@", 1); length_i, offset_i = int(length), int(offset); cursor[uri] = offset_i + length_i; return f"{length_i}@{offset_i}"
        length_i = int(s)
        prev = cursor.get(uri, 0)
        cursor[uri] = prev + length_i
        return f"{length_i}@{prev}"
    '''_mergefiles'''
    def _mergefiles(self, global_init_path: Optional[str], seg_paths: List[str], output_path: str) -> None:
        tmp_out = output_path + ".part"
        with open(tmp_out, "wb") as out:
            if global_init_path and self._fileok(global_init_path):
                with open(global_init_path, "rb") as fp: shutil.copyfileobj(fp, out, length=1024 * 1024)
            for p in seg_paths:
                with open(p, "rb") as fp: shutil.copyfileobj(fp, out, length=1024 * 1024)
        os.replace(tmp_out, output_path)
    '''_safenamefromurl'''
    def _safenamefromurl(self, url: str, max_len: int = 20) -> str:
        return hashlib.sha256(url.encode("utf-8")).hexdigest()[:max_len]
    '''_fileok'''
    def _fileok(self, path: str) -> bool:
        return os.path.exists(path) and os.path.getsize(path) > 0
    '''_atomicwrite'''
    def _atomicwrite(self, path: str, data: bytes) -> None:
        touchdir(os.path.dirname(os.path.abspath(path)) or ".")
        pid, tid = os.getpid(), threading.get_ident()
        tmp, last = f"{path}.tmp.{pid}.{tid}.{time.time_ns()}", None
        with open(tmp, "wb") as fp:
            fp.write(data)
            try: fp.flush(); os.fsync(fp.fileno())
            except Exception: pass
        for i in range(12):
            try: os.replace(tmp, path); return
            except PermissionError as e: last = e; time.sleep(min(0.5, 0.03 * (2 ** i)))
            except OSError as e: last = e; time.sleep(min(0.5, 0.03 * (2 ** i)))
        try:
            if os.path.exists(tmp): os.remove(tmp)
        except Exception:
            pass
        raise last


'''
Function:
    Implementation of JBSouMusicClient: https://www.jbsou.cn/
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import copy
from urllib.parse import urljoin
from rich.progress import Progress
from ..sources import BaseMusicClient
from ..utils import legalizestring, resp2json, usesearchheaderscookies, seconds2hms, extractdurationsecondsfromlrc, safeextractfromdict, cleanlrc, SongInfo


'''JBSouMusicClient'''
class JBSouMusicClient(BaseMusicClient):
    source = 'JBSouMusicClient'
    ALLOWED_SITES = ['netease', 'qq', 'kugou', 'kuwo', 'migu', 'qianqian'][:-2] # it seems qianqian and migu are useless, recorded in 2026-01-29
    def __init__(self, **kwargs):
        self.allowed_music_sources = list(set(kwargs.pop('allowed_music_sources', JBSouMusicClient.ALLOWED_SITES)))
        super(JBSouMusicClient, self).__init__(**kwargs)
        self.default_search_headers = {
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/144.0.0.0 Safari/537.36", 
            "accept": "application/json, text/javascript, */*; q=0.01", "accept-encoding": "gzip, deflate, br, zstd", "accept-language": "zh-CN,zh;q=0.9,en-US;q=0.8,en;q=0.7",
            "origin": "https://www.jbsou.cn", "x-requested-with": "XMLHttpRequest", "referer": "https://www.jbsou.cn/"
        }
        self.default_download_headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/142.0.0.0 Safari/537.36',
        }
        self.default_headers = self.default_search_headers
        self._initsession()
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        self.search_size_per_page = min(self.search_size_per_page, 10)
        allowed_music_sources = copy.deepcopy(self.allowed_music_sources)
        # construct search urls based on search rules
        base_url = 'https://www.jbsou.cn/'
        search_urls, page_size = [], self.search_size_per_page
        for source in JBSouMusicClient.ALLOWED_SITES:
            if source not in allowed_music_sources: continue
            source_default_rule, count = {'input': keyword, 'filter': 'name', 'type': source, 'page': 1}, 0
            source_default_rule.update(rule)
            while self.search_size_per_source > count:
                page_rule = copy.deepcopy(source_default_rule)
                page_rule['page'] = str(int(count // page_size) + 1)
                search_urls.append({'url': base_url, 'data': page_rule})
                count += page_size
        # return
        return search_urls
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: dict = None, request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides, base_url = request_overrides or {}, "https://www.jbsou.cn/"
        source = search_url['data']['type']
        # successful
        try:
            # --search results
            resp = self.post(**search_url, **request_overrides)
            resp.raise_for_status()
            search_results = resp2json(resp)['data']
            for search_result in search_results:
                # --download results
                if not isinstance(search_result, dict) or ('songid' not in search_result) or ('url' not in search_result): continue
                search_result['source'] = source
                song_info = SongInfo(source=self.source, root_source=search_result['source'])
                download_url = urljoin(base_url, search_result['url'])
                try: (resp := self.session.head(download_url, allow_redirects=True, **request_overrides)).raise_for_status(); download_url = resp.url
                except Exception: continue
                cover_url = urljoin(base_url, search_result.get('cover', "") or "")
                try: (resp := self.session.head(cover_url, timeout=10, allow_redirects=True, **request_overrides)).raise_for_status(); cover_url = resp.url
                except Exception: cover_url = cover_url
                song_info = SongInfo(
                    raw_data={'search': search_result, 'download': {}, 'lyric': {}}, source=self.source, song_name=legalizestring(safeextractfromdict(search_result, ['name'], None)),
                    singers=legalizestring(str(safeextractfromdict(search_result, ['artist'], "")).replace('/', ', ')), album=legalizestring(search_result.get('album')),
                    ext=download_url.split('?')[0].split('.')[-1], file_size='NULL', identifier=search_result['songid'], duration='-:-:-', lyric=None, cover_url=cover_url, 
                    download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides), root_source=search_result['source'],
                )
                if not song_info.with_valid_download_url: continue
                song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
                song_info.file_size = song_info.download_url_status['probe_status']['file_size']
                song_info.ext = song_info.download_url_status['probe_status']['ext'] if (song_info.download_url_status['probe_status']['ext'] and song_info.download_url_status['probe_status']['ext'] not in ('NULL', )) else song_info.ext
                # --lyric results
                try:
                    resp = self.get(urljoin(base_url, search_result['lrc']), **request_overrides)
                    resp.raise_for_status()
                    lyric, lyric_result = cleanlrc(resp.text), {'lyric': resp.text}
                    song_info.duration_s = extractdurationsecondsfromlrc(lyric)
                    song_info.duration = seconds2hms(song_info.duration_s)
                except:
                    lyric_result, lyric = dict(), 'NULL'
                song_info.lyric = lyric
                song_info.raw_data['lyric'] = lyric_result
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos


'''
Function:
    Implementation of SoundCloudMusicClient: https://soundcloud.com/discover
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import re
import copy
from .base import BaseMusicClient
from urllib.parse import urlencode
from rich.progress import Progress
from ..utils import legalizestring, resp2json, usesearchheaderscookies, seconds2hms, safeextractfromdict, SongInfo


'''SoundCloudMusicClient'''
class SoundCloudMusicClient(BaseMusicClient):
    source = 'SoundCloudMusicClient'
    def __init__(self, **kwargs):
        super(SoundCloudMusicClient, self).__init__(**kwargs)
        if self.default_cookies: assert ("oauth_token" in self.default_cookies), '"oauth_token" should be configured, refer to https://musicdl.readthedocs.io/zh/latest/Quickstart.html#soundcloud-music-download'
        self.client_id = None
        self.default_search_headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36 Edg/121.0.0.0",
        }
        self.default_download_headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36 Edg/121.0.0.0",
        }
        self.default_headers = self.default_search_headers
        if self.default_search_cookies: self.default_search_headers.update({'Authorization': self.default_cookies["oauth_token"]})
        if self.default_download_cookies: self.default_download_headers.update({'Authorization': self.default_cookies["oauth_token"]})
        self._initsession()
    '''_updateclientid'''
    def _updateclientid(self, request_overrides: dict = None):
        if self.client_id: return
        request_overrides = request_overrides or {}
        try: resp = self.session.get('https://soundcloud.com/', **request_overrides); resp.raise_for_status()
        except: self.client_id = '9jZvetLfDs6An08euQgJ0lYlHkKdGFzV'; return
        script_urls = re.findall(r'<script[^>]+src="([^"]+)"', resp.text)
        for url in reversed(script_urls):
            try: resp = self.session.get(url, **request_overrides); m = re.search(r'client_id\s*:\s*"([0-9a-zA-Z]{32})"', resp.text) if resp.status_code == 200 else None
            except Exception: continue
            if m: self.client_id = m.group(1); return
        self.client_id = '9jZvetLfDs6An08euQgJ0lYlHkKdGFzV'; return
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        self._updateclientid(request_overrides=request_overrides)
        # search rules
        default_rule = {
            'q': keyword, 'sc_a_id': 'ab15798461680579b387acf67441b40149e528cd', 'facet': 'genre', 'user_id': '704923-225181-486085-807554', 'client_id': self.client_id,
            'limit': '20', 'offset': '0', 'linked_partitioning': '1', 'app_version': '1769771069', 'app_locale': 'en',
        }
        default_rule.update(rule)
        # construct search urls based on search rules
        base_url = 'https://api-v2.soundcloud.com/search/tracks?'
        search_urls, page_size, count = [], self.search_size_per_page, 0
        while self.search_size_per_source > count:
            page_rule = copy.deepcopy(default_rule)
            page_rule['limit'] = page_size
            page_rule['offset'] = count
            search_urls.append(base_url + urlencode(page_rule))
            count += page_size
        # return
        return search_urls
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        guess_codec_func = lambda t: ((lambda preset, mime: "opus" if ("opus" in preset or "opus" in mime) else "aac" if ("aac" in preset or "mp4a" in mime or "audio/mp4" in mime or "m4a" in mime) else "mp3" if ("mp3" in preset or "audio/mpeg" in mime) else "abr" if ("abr" in preset) else "unknown")(
            (safeextractfromdict(t, ["preset"], "") or "").lower(), (safeextractfromdict(t, ["format", "mime_type"], "") or "").lower()
        ))
        guess_bitrate_kbps_func = lambda t: (lambda preset: (lambda m: int(m.group(1)) if m else 128 if preset == "mp3_0_1" else 64 if preset == "opus_0_0" else 128 if preset.startswith("abr") else 0)(re.search(r"(\d+)\s*k", preset)))(
            (safeextractfromdict(t, ["preset"], "") or "").lower()
        )
        quality_rank_func = lambda t: {"hq": 2, "sq": 1}.get((safeextractfromdict(t, ["quality"], "") or "").lower(), 0)
        codec_rank_func = lambda codec: {"opus": 4, "aac": 3, "abr": 2, "mp3": 1, "unknown": 0}.get((codec or "").lower(), 0)
        protocol_rank_func = lambda t: {"progressive": 2, "hls": 1}.get((safeextractfromdict(t, ["format", "protocol"], "") or "").lower(), 0)
        sort_key_func = lambda t: (lambda c, br: (quality_rank_func(t), br, codec_rank_func(c), protocol_rank_func(t)))(guess_codec_func(t), guess_bitrate_kbps_func(t))
        # successful
        try:
            # --search results
            resp = self.get(search_url, **request_overrides)
            resp.raise_for_status()
            search_results = resp2json(resp)['collection']
            for search_result in search_results:
                # --download results
                if not isinstance(search_result, dict) or ('id' not in search_result): continue
                for transcoding in sorted((safeextractfromdict(search_result, ['media', 'transcodings'], []) or []), key=sort_key_func, reverse=True):
                    if not isinstance(transcoding, dict): continue
                    preset, mime_type = transcoding.get('preset', '') or '', safeextractfromdict(transcoding, ['format', 'mime_type'], None) or ''
                    download_url, protocol = transcoding.get('url'), safeextractfromdict(transcoding, ['format', 'protocol'], None)
                    if str(protocol).startswith(('ctr-', 'cbc-')): continue # TODO: Solve DRM issues in SoundCloud
                    ext = (('opus' if ('opus' in preset or 'opus' in mime_type) else None) or ('m4a' if ('aac' in preset or 'm4a' in mime_type) else None) or 'mp3')
                    if f"{protocol}_{preset}" in {"original_download"}:
                        try: resp = self.get(f'https://api-v2.soundcloud.com/tracks/{search_result["id"]}/download', params={'client_id': self.client_id}, **request_overrides); resp.raise_for_status()
                        except Exception: continue
                        download_result = resp2json(resp=resp)
                        download_url = download_result.get('redirectUri')
                        if not download_url or not str(download_url).startswith('http'): continue
                    else:
                        try: resp = self.get(download_url, params={'client_id': self.client_id}, **request_overrides); resp.raise_for_status()
                        except Exception: continue
                        download_result = resp2json(resp=resp)
                        download_url = download_result.get('url')
                        if not download_url or not str(download_url).startswith('http'): continue
                    if str(protocol).lower() in {'hls'}:
                        try: resp = self.session.head(download_url, **request_overrides); resp.raise_for_status()
                        except Exception: continue
                        download_url_status = {'ok': True}
                    else:
                        download_url_status = self.audio_link_tester.test(download_url, request_overrides)
                    song_info = SongInfo(
                        raw_data={'search': search_result, 'download': download_result, 'lyric': {}}, source=self.source, song_name=legalizestring(safeextractfromdict(search_result, ['title'], None)),
                        singers=legalizestring(safeextractfromdict(search_result, ['publisher_metadata', 'artist'], None) or safeextractfromdict(search_result, ['user', 'username'], None)),
                        album=legalizestring(safeextractfromdict(search_result, ['publisher_metadata', 'album_title'], None)), ext=ext, file_size='NULL', identifier=search_result['id'],
                        duration_s=int(float(safeextractfromdict(search_result, ['duration'], 0)) / 1000), duration=seconds2hms(int(float(safeextractfromdict(search_result, ['duration'], 0)) / 1000)), 
                        lyric='NULL', cover_url=search_result.get('artwork_url'), download_url=download_url, download_url_status=download_url_status
                    )
                    if not song_info.with_valid_download_url: continue
                    if str(protocol).lower() in {'hls'}:
                        song_info.protocol, song_info.file_size = 'HLS', 'HLS'
                    else:
                        song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
                        song_info.file_size = song_info.download_url_status['probe_status']['file_size']
                        song_info.ext = song_info.download_url_status['probe_status']['ext'] if (song_info.download_url_status['probe_status']['ext'] and song_info.download_url_status['probe_status']['ext'] not in ('NULL', )) else song_info.ext
                    if song_info.with_valid_download_url: break
                if not song_info.with_valid_download_url: continue
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos


'''
Function:
    Implementation of KugouMusicClient: http://www.kugou.com/
Author:
    Zhenchao Jin
WeChat Official Account (微信公众号):
    Charles的皮卡丘
'''
import os
import copy
import time
import random
import base64
import hashlib
import warnings
import json_repair
from .base import BaseMusicClient
from urllib.parse import urlencode
from rich.progress import Progress
from pathvalidate import sanitize_filepath
from ..utils.hosts import KUGOU_MUSIC_HOSTS
from urllib.parse import urlparse, parse_qs
from ..utils.kugouutils import KugouMusicClientUtils, MUSIC_QUALITIES
from rich.progress import Progress, TextColumn, BarColumn, TimeRemainingColumn, MofNCompleteColumn
from ..utils import touchdir, legalizestring, byte2mb, resp2json, seconds2hms, usesearchheaderscookies, safeextractfromdict, optionalimport, useparseheaderscookies, obtainhostname, hostmatchessuffix, cleanlrc, SongInfo
warnings.filterwarnings('ignore')


'''KugouMusicClient'''
class KugouMusicClient(BaseMusicClient):
    source = 'KugouMusicClient'
    def __init__(self, **kwargs):
        super(KugouMusicClient, self).__init__(**kwargs)
        self.default_search_headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36'}
        self.default_parse_headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36'}
        self.default_download_headers = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36'}
        self.default_headers = self.default_search_headers
        self._initsession()
    '''_parsewithcggapi'''
    def _parsewithcggapi(self, file_hash: str, search_result: dict, request_overrides: dict = None):
        # init
        curl_cffi = optionalimport('curl_cffi')
        request_overrides = request_overrides or {}
        MUSIC_QUALITIES = ['lossless', 'exhigh', 'hires', 'standard', 'ogg']
        # safe fetch filesize func
        safe_fetch_filesize_func = lambda meta: (lambda s: (lambda: float(s))() if s.replace('.', '', 1).isdigit() else 0)(str(meta.get('size', '0.00MB')).removesuffix('MB').strip()) if isinstance(meta, dict) else 0
        # parse
        for quality in MUSIC_QUALITIES:
            try: resp = curl_cffi.requests.get(f"https://music-api2.cenguigui.cn/?kg=&id={file_hash}&type=song&format=json&level={quality}", timeout=10, impersonate="chrome131", verify=False, **request_overrides); resp.raise_for_status()
            except Exception: break
            download_result = json_repair.loads(resp.text)
            if 'data' not in download_result or (safe_fetch_filesize_func(download_result['data']) < 1): continue
            download_url = safeextractfromdict(download_result, ['data', 'url'], '')
            if not download_url or not str(download_url).startswith('http'): continue
            try: duration_in_secs = search_result.get('duration') or (float(search_result.get('timelen', 0) or 0) / 1000)
            except Exception: duration_in_secs = 0
            song_info = SongInfo(
                raw_data={'search': search_result, 'download': download_result, 'lyric': {}}, source=self.source, song_name=legalizestring(safeextractfromdict(download_result, ['data', 'name'], None)), singers=legalizestring(safeextractfromdict(download_result, ['data', 'artist'], None)), 
                album=legalizestring(search_result.get('album_name') or safeextractfromdict(search_result, ['albuminfo', 'name'], "")), ext=download_url.split('?')[0].split('.')[-1], file_size=str(safeextractfromdict(download_result, ['data', 'size'], "")).removesuffix('MB').strip() + ' MB',
                identifier=file_hash, duration_s=duration_in_secs, duration=seconds2hms(duration_in_secs), lyric='NULL', cover_url=safeextractfromdict(download_result, ['data', 'pic'], ""), download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
            )
            song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
            song_info.file_size = song_info.download_url_status['probe_status']['file_size']
            if song_info.with_valid_download_url: break
        # return
        return song_info
    '''_parsewiththirdpartapis'''
    def _parsewiththirdpartapis(self, file_hash: str, search_result: dict, request_overrides: dict = None):
        if self.default_cookies or request_overrides.get('cookies'): return SongInfo(source=self.source)
        for imp_func in [self._parsewithcggapi]:
            try:
                song_info_flac = imp_func(file_hash, search_result, request_overrides)
                if song_info_flac.with_valid_download_url: break
            except:
                song_info_flac = SongInfo(source=self.source)
        return song_info_flac
    '''_constructsearchurls'''
    def _constructsearchurls(self, keyword: str, rule: dict = None, request_overrides: dict = None):
        # init
        rule, request_overrides = rule or {}, request_overrides or {}
        # search rules
        default_rule = {"format": "json", "keyword": keyword, "showtype": 1, "page": 1, "pagesize": 10}
        default_rule.update(rule)
        # construct search urls based on search rules
        base_url = 'http://mobilecdn.kugou.com/api/v3/search/song?'
        search_urls, page_size, count = [], self.search_size_per_page, 0
        while self.search_size_per_source > count:
            page_rule = copy.deepcopy(default_rule)
            page_rule['pagesize'] = page_size
            page_rule['page'] = int(count // page_size) + 1
            search_urls.append(base_url + urlencode(page_rule))
            count += page_size
        # return
        return search_urls
    '''_search'''
    @usesearchheaderscookies
    def _search(self, keyword: str = '', search_url: str = '', request_overrides: dict = None, song_infos: list = [], progress: Progress = None, progress_id: int = 0):
        # init
        request_overrides = request_overrides or {}
        safe_fetch_filesize_func = lambda size: (lambda s: (lambda: float(s))() if s.replace('.', '', 1).isdigit() else 0)(size.removesuffix('MB').strip()) if isinstance(size, str) else 0
        # successful
        try:
            # --search results
            resp = self.get(search_url, **request_overrides)
            resp.raise_for_status()
            search_results = resp2json(resp)['data']['info']
            for search_result in search_results:
                # --download results
                if not isinstance(search_result, dict) or ('hash' not in search_result): continue
                song_info = SongInfo(source=self.source)
                song_info_flac = self._parsewiththirdpartapis(file_hash=search_result['hash'], search_result=search_result, request_overrides=request_overrides)
                for quality in MUSIC_QUALITIES:
                    if song_info_flac.with_valid_download_url and song_info_flac.ext in ('flac',): song_info = song_info_flac; break
                    try:
                        per_request_overrides = copy.deepcopy(request_overrides)
                        if 'impersonate' not in per_request_overrides and self.enable_curl_cffi: per_request_overrides['impersonate'] = random.choice(self.cc_impersonates)
                        self._autosetproxies()
                        per_request_overrides['proxies'] = per_request_overrides.pop('proxies', None) or self.session.proxies
                        download_result: dict = KugouMusicClientUtils.getsongurl(self.session, hash_value=search_result['hash'], quality=quality, request_overrides=per_request_overrides, cookies=copy.deepcopy(request_overrides.get('cookies') or self.default_cookies))
                        download_url = safeextractfromdict(download_result, ['url'], '') or safeextractfromdict(download_result, ['backupUrl'], '')
                    except:
                        download_result, download_url = {}, None
                    if not download_url:
                        md5_hex = hashlib.md5((search_result['hash'] + 'kgcloudv2').encode("utf-8")).hexdigest()
                        try:
                            resp = self.get(f"https://trackercdn.kugou.com/i/v2/?cdnBackup=1&behavior=download&pid=1&cmd=21&appid=1001&hash={search_result['hash']}&key={md5_hex}", **request_overrides)
                            resp.raise_for_status()
                            download_result: dict = resp2json(resp)
                            download_url = (
                                safeextractfromdict(download_result, ['url'], '') or safeextractfromdict(download_result, ['backup_url'], '') or safeextractfromdict(download_result, ['backupUrl'], '') or 
                                safeextractfromdict(download_result, ['mp3Url'], '') or safeextractfromdict(download_result, ['backupMp3Url'], '')
                            )
                        except:
                            continue
                    if download_url and isinstance(download_url, (list, tuple)): download_url = list(download_url)[0]
                    if not download_url or not str(download_url).startswith('http'): continue
                    song_info = SongInfo(
                        raw_data={'search': search_result, 'download': download_result, 'lyric': {}}, source=self.source, song_name=legalizestring(safeextractfromdict(search_result, ['songname'], None) or safeextractfromdict(search_result, ['songname_original'], None) or safeextractfromdict(search_result, ['filename'], None)),
                        singers=legalizestring(safeextractfromdict(search_result, ['singername'], None)), album=legalizestring(safeextractfromdict(search_result, ['album_name'], None)), ext=download_result.get('extName') or download_url.split('?')[0].split('.')[-1] or 'mp3', file_size_bytes=download_result.get('fileSize', 0),
                        file_size=byte2mb(download_result.get('fileSize', 0)), identifier=search_result['hash'], duration_s=safeextractfromdict(search_result, ['duration'], 0), duration=seconds2hms(search_result.get('duration')), lyric='NULL', cover_url=safeextractfromdict(search_result, ['trans_param', 'union_cover'], ""),
                        download_url=download_url, download_url_status=self.audio_link_tester.test(download_url, request_overrides),
                    )
                    song_info.download_url_status['probe_status'] = self.audio_link_tester.probe(song_info.download_url, request_overrides)
                    song_info.file_size = song_info.download_url_status['probe_status']['file_size']
                    song_info.ext = song_info.download_url_status['probe_status']['ext'] if (song_info.download_url_status['probe_status']['ext'] and song_info.download_url_status['probe_status']['ext'] != 'NULL') else song_info.ext
                    if song_info.cover_url and isinstance(song_info.cover_url, str): song_info.cover_url = song_info.cover_url.format(size=300)
                    if song_info_flac.with_valid_download_url and (safe_fetch_filesize_func(song_info.file_size) < safe_fetch_filesize_func(song_info_flac.file_size)): song_info = song_info_flac
                    if song_info.with_valid_download_url: break
                if not song_info.with_valid_download_url: song_info = song_info_flac
                if not song_info.with_valid_download_url: continue
                # --lyric results
                params = {'keyword': search_result.get('filename', ''), 'duration': search_result.get('duration', '99999'), 'hash': search_result['hash']}
                try:
                    resp = self.get('http://lyrics.kugou.com/search', params=params, **request_overrides)
                    resp.raise_for_status()
                    lyric_result = resp2json(resp=resp)
                    resp = self.get(f"http://lyrics.kugou.com/download?ver=1&client=pc&id={lyric_result['candidates'][0]['id']}&accesskey={lyric_result['candidates'][0]['accesskey']}&fmt=lrc&charset=utf8", **request_overrides)
                    resp.raise_for_status()
                    lyric_result['lyrics.kugou.com/download'] = resp2json(resp=resp)
                    lyric = lyric_result['lyrics.kugou.com/download']['content']
                    lyric = cleanlrc(base64.b64decode(lyric).decode('utf-8'))
                except:
                    lyric_result, lyric = dict(), 'NULL'
                song_info.raw_data['lyric'] = lyric_result if lyric_result else song_info.raw_data['lyric']
                song_info.lyric = lyric if (lyric and (lyric not in {'NULL'})) else song_info.lyric
                # --append to song_infos
                song_infos.append(song_info)
                # --judgement for search_size
                if self.strict_limit_search_size_per_page and len(song_infos) >= self.search_size_per_page: break
            # --update progress
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Success)")
        # failure
        except Exception as err:
            progress.update(progress_id, description=f"{self.source}.search >>> {search_url} (Error: {err})")
        # return
        return song_infos
    '''parseplaylist'''
    @useparseheaderscookies
    def parseplaylist(self, playlist_url: str, request_overrides: dict = None):
        request_overrides = request_overrides or {}
        playlist_url = self.session.head(playlist_url, allow_redirects=True, **request_overrides).url
        hostname = obtainhostname(url=playlist_url)
        if not hostname or not hostmatchessuffix(hostname, KUGOU_MUSIC_HOSTS): return []
        assert 'special/single/' in urlparse(playlist_url).path, 'kugou playlist link must look like "https://www.kugou.com/yy/special/single/6914288.html"'
        try: playlist_id = parse_qs(urlparse(playlist_url).query, keep_blank_values=False).get('id')[0]; assert playlist_id
        except: playlist_id = urlparse(playlist_url).path.strip('/').split('/')[-1].removesuffix('.html')
        headers = {'User-Agent': 'Android9-AndroidPhone-11239-18-0-playlist-wifi', 'Host': 'gatewayretry.kugou.com', 'x-router': 'pubsongscdn.kugou.com', 'mid': '239526275778893399526700786998289824956', 'dfid': '-', 'clienttime': str(time.time()).split('.')[0]}
        page, tracks, song_infos = 1, [], []
        while True:
            api_url = f'http://gatewayretry.kugou.com/v2/get_other_list_file?specialid={playlist_id}&need_sort=1&module=CloudMusic&clientver=11239&pagesize=300&specalidpgc={playlist_id}&userid=0&page={page}&type=0&area_code=1&appid=1005'
            kugou_signature_func = lambda api_url: hashlib.md5(("OIlwieks28dk2k092lksi2UIkp" + "".join(sorted(api_url.split("?", 1)[1].split("&"))) + "OIlwieks28dk2k092lksi2UIkp").encode("utf-8")).hexdigest()
            try: (resp := self.get(api_url + '&signature=' + kugou_signature_func(api_url), headers=headers, **request_overrides)).raise_for_status()
            except Exception: continue
            playlist_results = resp2json(resp=resp)
            if (not safeextractfromdict(playlist_results, ['data', 'info'], [])) or (float(safeextractfromdict(playlist_results, ['data', 'count'], 0)) <= len(tracks)): break
            tracks.extend(safeextractfromdict(playlist_results, ['data', 'info'], []))
        tracks = list({d["hash"]: d for d in tracks}.values())
        with Progress(TextColumn("{task.description}"), BarColumn(bar_width=None), MofNCompleteColumn(), TimeRemainingColumn(), refresh_per_second=10) as main_process_context:
            main_progress_id = main_process_context.add_task(f"{len(tracks)} songs found in playlist {playlist_id} >>> completed (0/{len(tracks)})", total=len(tracks))
            for idx, track_info in enumerate(tracks):
                if idx > 0: main_process_context.advance(main_progress_id, 1)
                main_process_context.update(main_progress_id, description=f"{len(tracks)} songs found in playlist {playlist_id} >>> completed ({idx}/{len(tracks)})")
                for third_part_api in [self._parsewithcggapi]:
                    try:
                        song_info = third_part_api(track_info['hash'], track_info, request_overrides=request_overrides)
                        if song_info.with_valid_download_url: song_infos.append(song_info); break
                    except:
                        continue
            main_process_context.advance(main_progress_id, 1)
            main_process_context.update(main_progress_id, description=f"{len(tracks)} songs found in playlist {playlist_id} >>> completed ({idx+1}/{len(tracks)})")
        song_infos = self._removeduplicates(song_infos=song_infos)
        work_dir = self._constructuniqueworkdir(keyword=playlist_id)
        for song_info in song_infos:
            song_info.work_dir = work_dir; episodes = song_info.episodes if isinstance(song_info.episodes, list) else []
            for eps_info in episodes: eps_info.work_dir = sanitize_filepath(os.path.join(work_dir, song_info.song_name)); touchdir(work_dir)
        return song_infos